import os
import glob
import xml.etree.ElementTree as ET
from datetime import datetime, timedelta
import numpy as np
from typing import List, Tuple, Dict, Any

# import from package
from resistics.ioHandlers.dataReader import DataReader
from resistics.dataObjects.timeData import TimeData
from resistics.utilities.utilsChecks import isMagnetic, isElectric, consistentChans
from resistics.utilities.utilsPrint import blockPrint
from resistics.utilities.utilsClean import (
    removeZeros,
    removeZerosSingle,
    removeNansSingle,
)


class DataReaderSPAM(DataReader):
    """Data reader for SPAM data

    SPAM data has the following characteristics:

    - SPAM raw data is single precision floats with unit Volts. 
    - Getting unscaled samples returns data with unit mV for both the electric and magnetic fields. This is because gain is removed in unscaled samples to ensure consistency when a single recording is made up of multiple data files, each with different gain settings
    - The start time in XTR files is the time of the first sample in the data
    - The end time in XTR files is the time of the last sample in the data

    In situations where a SPAM dataset is recorded in multiple small files, it is required that the recording is continuous. 

    Attributes
    ----------
    recChannels : Dict
        Channels in each data file
    dtype : np.float32
        The data type
    numHeaderFiles : int
        The number of header files
    numDataFiles : int
        The number of data files

    Methods
    -------
    __init__(dataPath)
        Initialise with path to the data directory
    setParameters()
        Set parameters specific to a data format
    getUnscaledSamples(**kwargs)
        Get raw, unscaled data
    getPhysicalSamples(**kwargs)
        Get data in physical units
    spamHeaders()
        Get sections and section headers to be read in for SPAM data
    chanDefaults()
        Get defaults values for channel headers
    readHeader()
        Read SPAM header files
    readHeaderXTR(headerFile)
        Read a XTR header file
    readHeaderXTRX(headerFile)
        Read a XTRX header files
    headersFromRawFile(rawFile, headers)
        Read headers from the data files
    mergeHeaders(headersList, chanHeadersList)
        Merge the headers from all the data files
    printDataFileList()
        Get data file information as a list of strings
    printDataFiles()
        Print data file information to terminal

    Notes
    -----
    Getting unscaled samples for SPAM data removes the gain rather than return exactly the values in the data files. In cases where there are multiple data files, it is not necessary that they have been recorded with the same gain. Therefore, to ensure consistency when looking at raw data, the gain is removed at the getUnscaledSamples stage rather than getPhysicalSamples, where it would have probably been more appropriate. This means that getUnscaledSamples returns data where all channels are in mV.

    The scalings to convert the raw data to mV are stored in the ts_lsb chan header and calculated out as the header files are being read.

    .. todo::
        Implement reading of XTRX header files
    """

    def setParameters(self) -> None:
        """Set some data reader parameters for reading SPAM data"""

        # get a list of the header and data files in the folder
        self.headerF = glob.glob(os.path.join(self.dataPath, "*.XTR"))
        if len(self.headerF) == 0:
            self.headerF = glob.glob(os.path.join(self.dataPath, "*.XTRX"))
        self.dataF = glob.glob(os.path.join(self.dataPath, "*.RAW"))
        # data byte information might be different for each file
        # so it is a dictionary
        self.dataByteOffset: Dict = {}
        self.recChannels = {}
        self.dataByteSize = 4
        # data type
        self.dtype = np.float32
        # get the number of data files and header files - this should be equal
        self.numHeaderFiles: int = len(self.headerF)
        self.numDataFiles: int = len(self.dataF)

    def getUnscaledSamples(self, **kwargs) -> TimeData:
        """Get raw data from data file, returned in mV

        SPAM raw data is single precision float with unit Volts. Calling this applies the ts_lsb calculated when the headers are read. This is because when a recording consists of multiple data files, each channel of each data file might have a different scaling. The only way to make the data consistent is to apply the ts_lsb scaling.  
        
        Therefore, this method returns the data in mV for all channels.

        Parameters
        ----------
        chans : List[str], optional
            List of channels to return if not all are required
        startSample : int, optional
            First sample to return
        endSample : int, optional
            Last sample to return

        Returns
        -------
        TimeData
            Time data object 
        """

        # initialise chans, startSample and endSample with the whole dataset
        options = self.parseGetDataKeywords(kwargs)

        # get the files to read and the samples to take from them, in the correct order
        dataFilesToRead, samplesToRead, scalings = self.getDataFilesForSamples(
            options["startSample"], options["endSample"]
        )
        numSamples = options["endSample"] - options["startSample"] + 1
        # set up the dictionary to hold the data
        data = {}
        for chan in options["chans"]:
            data[chan] = np.zeros(shape=(numSamples), dtype=self.dtype)

        # loop through chans and get data
        sampleCounter = 0
        for dFile, sToRead, scalar in zip(dataFilesToRead, samplesToRead, scalings):
            # get samples - this is inclusive
            dSamples = sToRead[1] - sToRead[0] + 1
            dSamplesRead = (
                dSamples * self.recChannels[dFile]
            )  # because spam files always record 5 channels
            # read the data
            byteOff = (
                self.dataByteOffset[dFile]
                + sToRead[0] * self.recChannels[dFile] * self.dataByteSize
            )
            dFilePath = os.path.join(self.dataPath, dFile)
            dataRead = np.memmap(
                dFilePath,
                dtype=self.dtype,
                mode="r",
                offset=byteOff,
                shape=(dSamplesRead),
            )
            # now need to unpack this
            for chan in options["chans"]:
                # check to make sure channel exists
                self.checkChan(chan)
                # get the channel index - the chanIndex should give the right order in the data file
                # as it is the same order as in the header file
                chanIndex = self.chanMap[chan]
                # use the range sampleCounter -> sampleCounter +  dSamples, because this actually means sampleCounter + dSamples - 1 as python ranges are not inclusive of the end value
                # scale by the lsb scalar here - note that these can be different for each file in the run
                data[chan][sampleCounter : sampleCounter + dSamples] = (
                    dataRead[chanIndex : dSamplesRead : self.recChannels[dFile]]
                    * scalar[chan]
                )
            # increment sample counter
            sampleCounter = sampleCounter + dSamples  # get ready for the next data read

        # return data
        startTime, stopTime = self.sample2time(
            options["startSample"], options["endSample"]
        )
        comments = []
        comments.append(
            "Unscaled data {} to {} read in from measurement {}, samples {} to {}".format(
                startTime,
                stopTime,
                self.dataPath,
                options["startSample"],
                options["endSample"],
            )
        )
        comments.append("Data read from {} files in total".format(len(dataFilesToRead)))
        comments.append(
            "Data scaled to mV for all channels using scalings in header files"
        )
        comments.append("Sampling frequency {}".format(self.getSampleFreq()))
        return TimeData(
            sampleFreq=self.getSampleFreq(),
            startTime=startTime,
            stopTime=stopTime,
            data=data,
            comments=comments,
        )

    def getDataFilesForSamples(
        self, startSample, endSample
    ) -> Tuple[List[str], List[List[int]], List[float]]:
        """Get the data files that have to be read to cover the sample range 
        
        Parameters
        ----------
        startSample : int
            Starting sample of the sample range
        endSamples : int
            Ending sample of the sample range

        Returns
        -------
        dataFilesToRead
            Time data object
        """

        # have the datafiles saved in sample order beginning with the earliest first
        # go through each datafile and find the range to be read
        dataFilesToRead = []
        samplesToRead = []
        scalings = []
        for idx, dFile in enumerate(self.dataFileList):
            fileStartSamp = self.dataRanges[idx][0]
            fileEndSamp = self.dataRanges[idx][1]
            if fileStartSamp > endSample or fileEndSamp < startSample:
                continue  # nothing to read from this file
            # in this case, there is some overlap with the samples to read
            dataFilesToRead.append(dFile)
            readFrom = 0  # i.e. the first sample in the datafile
            readTo = fileEndSamp - fileStartSamp  # this the last sample in the file
            if fileStartSamp < startSample:
                readFrom = startSample - fileStartSamp
            if fileEndSamp > endSample:
                readTo = endSample - fileStartSamp
            # this is an inclusive number readFrom to readTo including readTo
            samplesToRead.append([readFrom, readTo])
            scalings.append(self.scalings[idx])
        return dataFilesToRead, samplesToRead, scalings

    def getPhysicalSamples(self, **kwargs):
        """Get data scaled to physical values
        
        resistics uses field units, meaning physical samples will return the following:

        - Electrical channels in mV/km
        - Magnetic channels in mV
        - To get magnetic fields in nT, calibration needs to be performed

        Notes
        -----
        The method getUnscaledSamples multiplies the raw data by the ts_lsb converting it to mV. Because gain is removed when getting the unscaledSamples and all channel data is in mV, the only calculation that has to be done is to divide by the dipole lengths (east-west spacing and north-south spacing).
        
        To get magnetic fields in nT, they have to be calibrated.

        Parameters
        ----------
        chans : List[str]
            List of channels to return if not all are required
        startSample : int
            First sample to return
        endSample : int
            Last sample to return
        remaverage : bool
            Remove average from the data
        remzeros : bool
            Remove zeroes from the data
        remnans: bool
            Remove NanNs from the data

        Returns
        -------
        TimeData
            Time data object
        """

        # initialise chans, startSample and endSample with the whole dataset
        options = self.parseGetDataKeywords(kwargs)
        # get data
        timeData = self.getUnscaledSamples(
            chans=options["chans"],
            startSample=options["startSample"],
            endSample=options["endSample"],
        )
        # Ais applied in getUnscaledSamples to convert to mV - this is for ease of calculation and because each data file in the run might have a separate scaling
        # so all that is left is to divide by the dipole length in km and remove the average
        for chan in options["chans"]:
            if chan == "Ex":
                # multiply by 1000/self.getChanDx same as dividing by dist in km
                timeData.data[chan] = 1000 * timeData.data[chan] / self.getChanDx(chan)
                timeData.addComment(
                    "Dividing channel {} by electrode distance {} km to give mV/km".format(
                        chan, self.getChanDx(chan) / 1000.0
                    )
                )
            if chan == "Ey":
                # multiply by 1000/self.getChanDy same as dividing by dist in km
                timeData.data[chan] = 1000 * timeData.data[chan] / self.getChanDy(chan)
                timeData.addComment(
                    "Dividing channel {} by electrode distance {} km to give mV/km".format(
                        chan, self.getChanDy(chan) / 1000.0
                    )
                )

            # if remove zeros - False by default
            if options["remzeros"]:
                timeData.data[chan] = removeZerosSingle(timeData.data[chan])
            # if remove nans - False by default
            if options["remnans"]:
                timeData.data[chan] = removeNansSingle(timeData.data[chan])
            # remove the average from the data - True by default
            if options["remaverage"]:
                timeData.data[chan] = timeData.data[chan] - np.average(
                    timeData.data[chan]
                )

        # add comments
        timeData.addComment(
            "Remove zeros: {}, remove nans: {}, remove average: {}".format(
                options["remzeros"], options["remnans"], options["remaverage"]
            )
        )
        return timeData

    def spamHeaders(self) -> Tuple[List[str], Dict[str, str]]:
        """Get the sections in SPAM header files (XTR and XTRX)

        Returns
        -------
        sections : List[str]
            The sections in the header files
        sectionHeaders : Dict[str, str]
            The headers in each section to be read in 
        """

        sections = ["STATUS", "TITLE", "PROJECT", "FILE", "SITE", "CHANNAME", "DATA"]
        sectionHeaders = {}
        sectionHeaders["STATUS"] = ["STATUS"]
        sectionHeaders["TITLE"] = ["AUTHOR", "VERSION", "DATE", "COMMENT"]
        sectionHeaders["FILE"] = ["NAME", "FREQBAND", "DATE"]
        sectionHeaders["CHANNAME"] = ["ITEMS", "NAME"]
        sectionHeaders["DATA"] = ["ITEMS", "CHAN"]
        return sections, sectionHeaders

    def chanDefaults(self) -> Dict[str, Any]:
        """Get defaults for channel headers

        Returns
        -------
        Dict[str, Any]
            Dictionary of headers for channels and default values
        """

        chanH = {}
        chanH["gain_stage1"] = 1
        chanH["gain_stage2"] = 1
        chanH["hchopper"] = 0  # this depends on sample frequency
        chanH["echopper"] = 0
        # channel output information (sensor_type, channel_type, ts_lsb, pos_x1, pos_x2, pos_y1, pos_y2, pos_z1, pos_z2, sensor_sernum)
        chanH["ats_data_file"] = ""
        chanH["num_samples"] = 0
        chanH["sensor_type"] = ""
        chanH["channel_type"] = ""
        chanH["ts_lsb"] = 1
        # the lsb/scaling is not applied. data is raw voltage which needs to be scaled
        # an lsb is constructed from the scaling in the XTR/XTRX file to take the data to mV
        chanH["scaling_applied"] = False  # check this
        chanH["pos_x1"] = 0
        chanH["pos_x2"] = 0
        chanH["pos_y1"] = 0
        chanH["pos_y2"] = 0
        chanH["pos_z1"] = 0
        chanH["pos_z2"] = 0
        chanH["sensor_sernum"] = 0
        return chanH

    def readHeader(self) -> None:
        """Read header files

        For SPAM data, the may be more than one header file as data can be split up into smaller files as it is recorded. In that case, the header information should be somehow merged.
    
        All sampling frequencies should be the same
        """

        # read header files
        self.headersList = []
        self.chanHeadersList = []
        for headerFile in self.headerF:
            if "xtrx" in headerFile.lower():
                headers, chanHeaders = self.readHeaderXTRX(headerFile)
            else:
                headers, chanHeaders = self.readHeaderXTR(headerFile)
            self.headersList.append(headers)
            self.chanHeadersList.append(chanHeaders)

        # check to make sure no gaps
        # calculate out the sample ranges
        # and list the data files for each sample
        self.mergeHeaders(self.headersList, self.chanHeadersList)

    def readHeaderXTR(self, headerFile: str) -> None:
        """Read a XTR header file

        The raw data for SPAM is in single precision Volts. However, if there are multiple data files for a single recording, each one may have a different gain. Therefore, a scaling has to be calculated for each data file and channel. This scaling will convert all channels to mV. 

        For the most part, this method only reads recording information. However, it does additionally calculate out the lsb scaling and store it in the ts_lsb channel header. More information is provided in the notes.

        Notes
        -----
        The raw data for SPAM is in single precision floats and record the raw Voltage measurements of the sensors. However, if there are multiple data files for a single continuous recording, each one may have a different gain. Therefore, a scaling has to be calculated for each data file. 

        For electric channels, the scaling begins with the scaling provided in the header file in the DATA section. This incorporates any gain occuring in the device. This scaling is further amended by a conversion to mV and polarity reversal,

        .. code-block:: text
        
            scaling = read scaling from DATA section of header file
            scaling = 1000 * scaling , 
            scaling = -1000 * scaling ,
            ts_lsb = scaling ,
        
        where the reason for the 1000 factor in line 2 is not clear, nor is the polarity reversal. However, this information was provided by people more familiar with the data format.
        
        For magnetic channels, the scaling in the header file DATA section is ignored. This is because it includes a static gain correction, which would be duplicated at the calibration stage. Therefore, this is not included at this point.

        .. code-block:: text 
        
            scaling = -1000 ,
            ts_lsb = scaling ,
        
        This scaling converts the magnetic data from V to mV.

        Parameters
        ----------
        headerFile : str
            The XTR header file to read in
        """

        with open(headerFile, "r") as f:
            lines = f.readlines()
        sectionLines = {}
        # let's get data
        for line in lines:
            line = line.strip()
            line = line.replace("'", " ")
            # continue if line is empty
            if line == "":
                continue
            if "[" in line:
                sec = line[1:-1]
                sectionLines[sec] = []
            else:
                sectionLines[sec].append(line)
        # the base class is built around a set of headers based on ATS headers
        # though this is a bit more work here, it saves lots of code repetition
        headers = {}
        # recording information (start_time, start_date, stop_time, stop_date, ats_data_file)
        fileLine = sectionLines["FILE"][0]
        fileSplit = fileLine.split()
        headers["sample_freq"] = np.absolute(float(fileSplit[-1]))
        timeLine = sectionLines["FILE"][2]
        timeSplit = timeLine.split()
        # these are the unix time stamps
        startDate = float(timeSplit[1] + "." + timeSplit[2])
        datetimeStart = datetime.utcfromtimestamp(startDate)
        stopDate = float(timeSplit[3] + "." + timeSplit[4])
        datetimeStop = datetime.utcfromtimestamp(stopDate)
        headers["start_date"] = datetimeStart.strftime("%Y-%m-%d")
        headers["start_time"] = datetimeStart.strftime("%H:%M:%S.%f")
        headers["stop_date"] = datetimeStop.strftime("%Y-%m-%d")
        headers["stop_time"] = datetimeStop.strftime("%H:%M:%S.%f")
        # here calculate number of samples
        deltaSeconds = (datetimeStop - datetimeStart).total_seconds()
        # calculate number of samples - have to add one because the time given in SPAM recording is the actual time of the last sample
        numSamples = int(deltaSeconds * headers["sample_freq"]) + 1
        # put these in headers for ease of future calculations in merge headers
        headers["num_samples"] = numSamples
        # spam datasets only have the one data file for all channels
        headers["ats_data_file"] = fileSplit[1]
        # data information (meas_channels, sample_freq)
        chanLine = sectionLines["CHANNAME"][0]
        # this gets reformatted to an int later
        headers["meas_channels"] = chanLine.split()[1]
        numChansInt = int(headers["meas_channels"])
        # deal with the channel headers
        chanHeaders = []
        for iChan in range(0, numChansInt):
            chanH = self.chanDefaults()
            # set the sample frequency from the main headers
            chanH["sample_freq"] = headers["sample_freq"]
            # line data - read through the data in the correct channel order
            chanLine = sectionLines["CHANNAME"][iChan + 1]
            chanSplit = chanLine.split()
            dataLine = sectionLines["DATA"][iChan + 1]
            dataSplit = dataLine.split()
            # channel input information (gain_stage1, gain_stage2, hchopper, echopper)
            chanH["gain_stage1"] = 1
            chanH["gain_stage2"] = 1
            # channel output information (sensor_type, channel_type, ts_lsb, pos_x1, pos_x2, pos_y1, pos_y2, pos_z1, pos_z2, sensor_sernum)
            chanH["ats_data_file"] = fileSplit[1]
            chanH["num_samples"] = numSamples

            # channel information
            # spams often use Bx, By - use H within the software as a whole
            chanH["channel_type"] = consistentChans(chanSplit[2])
            # the sensor number is a bit of a hack - want MFSXXe or something - add MFS in front of the sensor number - this is liable to break
            # at the same time, set the chopper
            calLine = sectionLines["200{}003".format(iChan + 1)][0]
            calSplit = calLine.split()
            if isMagnetic(chanH["channel_type"]):
                chanH["sensor_sernum"] = calSplit[
                    2
                ]  # the last three digits is the serial number
                sensorType = calSplit[1].split("_")[1][-2:]
                chanH["sensor_type"] = "MFS{:02d}".format(int(sensorType))
                if "LF" in calSplit[1]:
                    chanH["hchopper"] = 1
            else:
                chanH["sensor_type"] = "ELC00"
                if "LF" in calLine:
                    chanH["echopper"] = 1

            # data is raw voltage of sensors
            # both E and H fields need polarity reversal (from email with Reinhard)
            # get scaling from headers
            scaling = float(dataSplit[-2])
            if isElectric(chanH["channel_type"]):
                # the factor of 1000 is not entirely clear
                lsb = 1000.0 * scaling
                # volts to millivolts and a minus to switch polarity giving data in mV
                lsb = -1000.0 * lsb
            else:
                # volts to millivolts and a minus to switch polarity giving data in mV
                # scaling in header file is ignored because it duplicates static gain correction in calibration
                lsb = -1000.0
            chanH["ts_lsb"] = lsb

            # the distances
            if chanSplit[2] == "Ex":
                chanH["pos_x1"] = float(dataSplit[4]) / 2
                chanH["pos_x2"] = chanH["pos_x1"]
            if chanSplit[2] == "Ey":
                chanH["pos_y1"] = float(dataSplit[4]) / 2
                chanH["pos_y2"] = chanH["pos_y1"]
            if chanSplit[2] == "Ez":
                chanH["pos_z1"] = float(dataSplit[4]) / 2
                chanH["pos_z2"] = chanH["pos_z1"]

            # append chanHeaders to the list
            chanHeaders.append(chanH)

        # check information from raw file headers
        self.headersFromRawFile(headers["ats_data_file"], headers)
        # return the headers and chanHeaders from this file
        return headers, chanHeaders

    def readHeaderXTRX(self, headerFile):
        """Read a XTRX header files

        XTRX are newer header files and will supercede XTR

        Parameters
        ----------
        headerFile : str
            The XTRX header file to read in
        """

        raise NotImplementedError("Support for XTRX files has not yet been implemented")

    def headersFromRawFile(self, rawFile: str, headers: Dict) -> None:
        """Read headers from the raw data files
        
        Read the headers from the raw file and figure out the data byte offset.     

        Parameters
        ----------
        rawFile : str
            The .RAW data file
        headers : Dict
            A headers dictionary

        Notes
        -----
        Open with encoding ISO-8859-1 because it has a value for all bytes unlike other encoding. In particular, want to find number of samples and the size of the header. The extended header is ignored.
        """

        dFile = open(os.path.join(self.dataPath, rawFile), "r", encoding="ISO-8859-1")
        generalHeaderString = dFile.read(1000)  # this should be long enough
        generalSplit = generalHeaderString.split()
        # read GENERAL HEADER
        generalHeader = {}
        generalHeader["recLength"] = int(generalSplit[0])
        generalHeader["fileType"] = generalSplit[1]
        generalHeader["wordLength"] = int(generalSplit[2])
        generalHeader["version"] = generalSplit[3]
        generalHeader["procId"] = generalSplit[4]
        generalHeader["numCh"] = int(generalSplit[5])
        generalHeader["totalRec"] = int(generalSplit[6])
        generalHeader["firstEvent"] = int(generalSplit[7])
        generalHeader["numEvent"] = int(generalSplit[8])
        generalHeader["extend"] = int(generalSplit[9])

        # read EVENT HEADER - there can be multiple of these, but normally only the one
        # Multiple events are largely deprecated. Only a single event is used
        eventHeaders = []
        fileSize = os.path.getsize(os.path.join(self.dataPath, rawFile))
        record = generalHeader["firstEvent"]
        for ir in range(0, generalHeader["numEvent"]):
            seekPt = (record - 1) * generalHeader["recLength"]
            if not seekPt > fileSize:
                # seek from beginning of file
                dFile.seek(seekPt, 0)
                # read extra to make sure
                eventString = dFile.read(1000)
                eventSplit = eventString.split()
                eH = {}
                eH["start"] = int(eventSplit[0])
                eH["startms"] = int(eventSplit[1])
                eH["stop"] = int(eventSplit[2])
                eH["stopms"] = int(eventSplit[3])
                eH["cvalue1"] = float(eventSplit[4])
                eH["cvalue2"] = float(eventSplit[5])
                eH["cvalue3"] = float(eventSplit[6])
                eH["EHInfile"] = int(eventSplit[7])
                eH["nextEH"] = int(eventSplit[8])
                eH["previousEH"] = int(eventSplit[9])
                eH["numData"] = int(eventSplit[10])
                eH["startData"] = int(eventSplit[11])
                eH["extended"] = int(eventSplit[12])
                eventHeaders.append(eH)
                if eH["nextEH"] < generalHeader["totalRec"]:
                    record = eH["nextEH"]  # set to go to next eH
                else:
                    break  # otherwise break out of for loops
        # close the data file
        dFile.close()
        # now compare number of samples with that calculated previously
        if eventHeaders[0]["numData"] != headers["num_samples"]:
            self.printWarning("Data file: {}".format(dFile))
            self.printWarning(
                "Number of samples in raw file header {} does not equal that calculated from data {}".format(
                    eventHeaders[0]["numData"], headers["num_samples"]
                )
            )
            self.printWarning("Number of samples calculated from data will be used")
        # set the byte offset for the file
        self.dataByteOffset[rawFile] = (
            eventHeaders[0]["startData"] - 1
        ) * generalHeader["recLength"]
        self.recChannels[rawFile] = generalHeader["numCh"]

    def mergeHeaders(self, headersList: List, chanHeadersList: List) -> None:
        """Merge headers from all the header files

        Checks all the header files to see if there are any gaps and calculates the sample ranges for each file together with the total number of samples. Sets the start and end time of the recording and class variables datetimeStart and datetimeStop.        

        Parameters
        ----------
        headersList : List
            List of headers from each data file
        chanHeadersList :
            List of chan headers from each data file
        """

        # take the first header as an example
        self.headers = headersList[0]
        self.chanHeaders = chanHeadersList[0]
        if len(headersList) == 1:
            # just fill in the data file list and data ranges
            self.dataFileList = [self.headers["ats_data_file"]]
            self.dataRanges = [[0, self.headers["num_samples"] - 1]]
            self.scalings = []
            tmp = {}
            for cHeader in self.chanHeaders:
                tmp[cHeader["channel_type"]] = cHeader["ts_lsb"]
            self.scalings.append(tmp)
            return  # then there was only one file - no need to do all the below

        # make sure that all headers have the same sample rate
        # and save the start and stop times and dates
        startTimes = []
        stopTimes = []
        numSamples = []
        for idx, header in enumerate(headersList):
            if header["sample_freq"] != self.headers["sample_freq"]:
                self.printWarning(
                    "Not all datasets in {} have the same sample frequency".format(
                        self.dataPath
                    )
                )
                self.printWarning("Exiting")
                exit()
            if header["meas_channels"] != self.headers["meas_channels"]:
                self.printWarning(
                    "Not all datasets in {} have the same number of channels".format(
                        self.dataPath
                    )
                )
                self.printWarning("Exiting")
                exit()
            # now store startTimes, stopTimes and numSamples
            # do this as datetimes, will be easier
            startString = "{} {}".format(header["start_date"], header["start_time"])
            stopString = "{} {}".format(header["stop_date"], header["stop_time"])
            datetimeStart = datetime.strptime(startString, "%Y-%m-%d %H:%M:%S.%f")
            datetimeStop = datetime.strptime(stopString, "%Y-%m-%d %H:%M:%S.%f")
            startTimes.append(datetimeStart)
            stopTimes.append(datetimeStop)
            numSamples.append(header["num_samples"])
        # check the start and end times
        sampleTime = timedelta(seconds=1.0 / self.headers["sample_freq"])
        # sort by start times
        sortIndices = sorted(list(range(len(startTimes))), key=lambda k: startTimes[k])
        # now sort stop times by the same indices
        check = True
        for i in range(1, self.numHeaderFiles):
            # get the stop time of the previous dataset
            stopTimePrev = stopTimes[sortIndices[i - 1]]
            startTimeNow = startTimes[sortIndices[i]]
            if startTimeNow != stopTimePrev + sampleTime:
                self.printWarning(
                    "There is a gap between the datafiles in {}".format(self.dataPath)
                )
                self.printWarning(
                    "Please separate out datasets with gaps into separate folders"
                )
                # print out where the gap was found
                self.printWarning("Gap found between datafiles:")
                self.printWarning(
                    "1. {}".format(headersList[sortIndices[i - 1]]["ats_data_file"])
                )
                self.printWarning(
                    "2. {}".format(headersList[sortIndices[i]]["ats_data_file"])
                )
                # set check as false
                check = False
        # if did not pass check, then exit
        if not check:
            exit()

        # make sure there are no gaps
        totalSamples = sum(numSamples)

        # get a list of all the datafiles, scalings and the sample ranges
        self.dataFileList = []
        self.dataRanges = []
        self.scalings = []
        sample = -1
        # now need some sort of lookup table to say where the sample ranges are
        for i in range(0, self.numHeaderFiles):
            iSort = sortIndices[i]  # get the sorted index
            self.dataFileList.append(headersList[iSort]["ats_data_file"])
            startSample = sample + 1
            endSample = (
                startSample + numSamples[iSort] - 1
            )  # -1 because this is inclusive of the start sample
            self.dataRanges.append([startSample, endSample])
            # increment sample
            sample = endSample
            # save the scalings for each chan
            tmp = {}
            for cHeader in self.chanHeadersList[iSort]:
                tmp[cHeader["channel_type"]] = cHeader["ts_lsb"]
            self.scalings.append(tmp)

        # now set the LSB information for the chanHeaders
        # i.e. if they change, this should reflect that
        for i in range(0, len(self.chanHeaders)):
            chan = self.chanHeaders[i]["channel_type"]
            lsbSet = set()
            for scalar in self.scalings:
                lsbSet.add(scalar[chan])
            if len(lsbSet) == 1:
                self.chanHeaders[i]["ts_lsb"] = list(lsbSet)[0]
            else:
                self.printWarning(
                    "Multiple different LSB values found for chan {}: {}".format(
                        chan, list(lsbSet)
                    )
                )
                self.printWarning(
                    "This is handled, but the header information given will show only a single LSB value"
                )
                self.chanHeaders[i]["ts_lsb"] = list(lsbSet)[0]

        # set start and end time for headers and chan headers
        # do the same with number of samples
        datetimeStart = min(startTimes)
        datetimeStop = max(stopTimes)
        self.headers["start_date"] = datetimeStart.strftime("%Y-%m-%d")
        self.headers["start_time"] = datetimeStart.strftime("%H:%M:%S.%f")
        self.headers["stop_date"] = datetimeStop.strftime("%Y-%m-%d")
        self.headers["stop_time"] = datetimeStop.strftime("%H:%M:%S.%f")
        self.headers["num_samples"] = totalSamples
        # set datafiles = the whole list of datafiles
        self.headers["ats_data_file"] = self.dataFileList
        for iChan in range(0, len(self.chanHeaders)):
            self.chanHeaders[iChan]["start_date"] = datetimeStart.strftime("%Y-%m-%d")
            self.chanHeaders[iChan]["start_time"] = datetimeStart.strftime(
                "%H:%M:%S.%f"
            )
            self.chanHeaders[iChan]["stop_date"] = datetimeStop.strftime("%Y-%m-%d")
            self.chanHeaders[iChan]["stop_time"] = datetimeStop.strftime("%H:%M:%S.%f")
            self.chanHeaders[iChan]["num_samples"] = totalSamples
            self.chanHeaders[iChan]["ats_data_file"] = self.dataFileList

    def printDataFileList(self) -> List[str]:
        """Information about the data files as a list of strings
        
        Returns
        -------
        List[str]
            List of information about the data files
        """

        textLst: List[str] = []
        textLst.append("Data File\t\tSample Ranges")
        for dFile, sRanges in zip(self.dataFileList, self.dataRanges):
            textLst.append("{}\t\t{} - {}".format(dFile, sRanges[0], sRanges[1]))
        textLst.append("Total samples = {}".format(self.getNumSamples()))
        return textLst

    def printDataFileInfo(self) -> None:
        """Print a list of the data files"""

        blockPrint(
            "{} Data File List".format(self.__class__.__name__),
            self.printDataFileList(),
        )




import os
import numpy as np
import scipy.signal as signal
import scipy.interpolate as interp
from typing import List

# import from package
from resistics.calculators.processorSingleSite import ProcessorSingleSite
from resistics.utilities.utilsSmooth import smooth1d
from resistics.utilities.utilsRobust import (
    chatterjeeMachler,
    sampleMAD0,
    hermitianTranspose,
    olsModel,
    mmestimateModel,
)


class ProcessorRemoteReference(ProcessorSingleSite):
    """Perform remote reference processing of magnetotelluric data

    Remote reference processing uses a different equation for spectral power than the single site processor. The equation that the processor uses is:

    Attributes
    ----------
    winSelector : WindowSelect
        A window selector object which defines which windows to use in the linear model
    decParams : DecimationParameters
        DecimationParameters object with information about the decimation scheme
    winParams : WindowParameters
        WindowParameters object with information about the windowing
    outpath : str 
        Location to put the calculated transfer functions (Edi files)
    inSite : str 
        The site to use for the input channels 
    inChannels: List[str] (["Hx", "Hy"])
        List of hannels to use as input channels for the linear system
    inSize : int 
        Number of input channels
    outSite : str 
        The site to use for the output channels
    outChannels : List[str] (["Ex", "Ey"])
        List of channels to use as output channels for the linear system
    outSize : int
        Number of output channels
    allChannels : List[str] 
        inChannels and outChannels combined into a single list
    crossChannels : List[str] 
        
    intercept : bool (default False)
        Flag for including an intercept (static) term in the linear system
    method : str (options, "ols", "cm") 
        String for describing what solution method to use
    win : str (default hanning)
        Window function to use in robust solution
    winSmooth : int (default -1)
        The size of the window smoother. If -1, this will be autocalculated based on data size
    postpend : str (default "")
        String to postpend to the output filename to help file management
    evalFreq : List[float] or np.ndarray
        The evaluation frequencies
    impedances : List

    variances : List

    Methods
    -------
    __init__(proj, maskData)
        Initialise with a Project instance and MaskData instance


    printList()
        Class status returned as list of strings
    """

    def __init__(self, winSelector, outpath: str):
        # required
        self.winSelector = winSelector
        self.decParams = winSelector.decParams
        self.winParams = winSelector.winParams
        self.outpath: str = outpath

        # default parameters for user options
        self.inSite: str = ""
        self.inChannels: List[str] = []
        self.outSite: str = ""
        self.outChannels: List[str] = []
        self.remoteSite: str = ""
        self.remoteChannels: List[str] = []
        # smoothing options
        self.win: str = "hanning"
        self.winSmooth: int = -1
        # intercept options
        self.intercept: bool = False
        # output filename
        self.postpend: str = ""
        # evaluation frequency data
        self.evalFreq = []
        self.impedances = []
        self.variances = []

    def setRemote(self, remoteSite: str, remoteChannels: List[str]) -> None:
        """Set information about input site and channels
    
        Parameters
        ----------
        remoteSite : str
            Site to use for input channel data
        remoteChannels : List[str]
            Channels to use as the remote reference in the linear system
        """

        self.remoteSite = remoteSite
        self.remoteChannels = remoteChannels
        self.remoteSize = len(remoteChannels)
        self.printText(
            "Remote reference set with site {} and channels {}".format(
                self.remoteSite, self.remoteChannels
            )
        )

    def process(self):
        """Process spectra data

        The processing sequence is as below:

        For each decimation level
            Get shared (unmasked) windows for all relevant sites (inSite and outSite)
            For shared unmasked windows
                Calculate out the spectral power data
                Interpolate calculated spectral power data to the evaluation frequencies for the decimation level 
            For each evaluation frequency
                Do the robust processing to calculate the transfer function at that evaluation frequency


        NOTES
        -----
        The spectral power data is smoothed as this tends to improve results. The smoothing can be changed by setting the smoothing parameters. This method is still subject to change in the future as it is an area of active work
        """

        numLevels = self.decParams.numLevels
        dataChans = self.inChannels + self.outChannels
        for iDec in range(0, numLevels):
            # print out some info
            self.printText("Processing decimation level {}".format(iDec))

            fs = self.winSelector.decParams.getSampleFreqLevel(iDec)
            # get the number of all shared windows and the number of unmasked windows
            # unmasked windows are ones that will actually be used in the calculation
            numWindows = self.winSelector.getNumSharedWindows(iDec)
            unmaskedWindows = self.winSelector.getUnmaskedWindowsLevel(iDec)
            numUnmasked = len(unmaskedWindows)
            self.printText(
                "Total shared windows for decimation level = {}".format(numWindows)
            )
            self.printText(
                "Total unmasked windows for decimation level = {}".format(numUnmasked)
            )
            if numUnmasked == 0:
                self.printText(
                    "No unmasked windows found at this decimation level ({:d}), continuing to next level".format(
                        iDec
                    )
                )
                continue  # continue to next decimation level
            self.printText("{} windows will be processed".format(numUnmasked))

            # get the evaluation frequencies
            evalFreq = self.decParams.getEvalFrequenciesForLevel(iDec)
            # set some variables
            totalChans = self.inSize + self.outSize
            numEvalFreq = len(evalFreq)
            dataSize = self.winSelector.getDataSize(iDec)
            freq = np.linspace(0, fs / 2, dataSize)
            # get the window smoothing params
            smoothLen = self.getWindowSmooth(datasize=dataSize)

            # create the data array
            # for each evaluation frequency
            # keep the spectral power information for all windows
            evalFreqData = np.empty(
                shape=(numEvalFreq, numWindows, totalChans, self.remoteSize),
                dtype="complex",
            )

            # an array for the in and out channels fourier data
            winDataArray = np.empty(shape=(totalChans, dataSize), dtype="complex")
            # an array for the remote reference fourier data
            winRemoteArray = np.empty(
                shape=(self.remoteSize, dataSize), dtype="complex"
            )
            # an array for the power spectra data
            winSpectraMatrix = np.empty(
                shape=(totalChans, self.remoteSize, dataSize), dtype="complex"
            )

            # loop over shared windows
            localWin = 0
            global2local = {}
            for iWin in unmaskedWindows:
                # do the local to global map
                global2local[iWin] = localWin

                # get the window for the input site
                inSF, inReader = self.winSelector.getSpecReaderForWindow(
                    self.inSite, iDec, iWin
                )
                inData = inReader.readBinaryWindowGlobal(iWin).data
                # get the window and channels for the output site
                if self.outSite != self.inSite:
                    outSF, outReader = self.winSelector.getSpecReaderForWindow(
                        self.outSite, iDec, iWin
                    )
                    outData = outReader.readBinaryWindowGlobal(iWin).data
                else:
                    outData = inData

                # now get the remote reference data - assume this does not equal input or output
                remoteSF, remoteReader = self.winSelector.getSpecReaderForWindow(
                    self.remoteSite, iDec, iWin
                )
                remoteData = remoteReader.readBinaryWindowGlobal(iWin).data

                # get data into the right part of the arrays
                for i in range(0, self.inSize):
                    winDataArray[i] = inData[self.inChannels[i]]
                for i in range(0, self.outSize):
                    winDataArray[self.inSize + i] = outData[self.outChannels[i]]
                for i in range(0, self.remoteSize):
                    winRemoteArray[i] = remoteData[self.remoteChannels[i]]

                # and now can fill the parts of the matrix
                # recall, smooth the power spectra
                for iD, dataChan in enumerate(dataChans):
                    for iR, remoteChan in enumerate(self.remoteChannels):
                        # calculate each one, cannot use complex symmetry here
                        # cannot use conjugate symmetry like with the single site processor
                        winSpectraMatrix[iD, iR] = smooth1d(
                            winDataArray[iD] * np.conjugate(winRemoteArray[iR]),
                            smoothLen,
                            self.win,
                        )
                # after running through all windows, calculate evaluation frequencies
                # calculate frequency array
                evalFreqData[:, localWin] = self.calcEvalFrequencyData(
                    freq, evalFreq, winSpectraMatrix
                )

                # increment local window
                localWin = localWin + 1

            # now all the data has been collected
            # for each evaluation frequency, do the robust processing
            # and get the evaluation frequency data
            for eIdx in range(0, numEvalFreq):
                self.printText(
                    "Processing evaluation frequency = {:.6f} [Hz], period = {:.6f} [s]".format(
                        evalFreq[eIdx], 1.0 / evalFreq[eIdx]
                    )
                )
                # get the constrained windows for the evaluation frequency
                evalFreqWindows = self.winSelector.getWindowsForFreq(iDec, eIdx)
                if len(evalFreqWindows) == 0:  # no windows meet constraints
                    self.printText("No windows found - possibly due to masking")
                    continue
                localWinIndices = []
                for iW in evalFreqWindows:
                    localWinIndices.append(global2local[iW])
                self.printText(
                    "{:d} windows will be solved for".format(len(localWinIndices))
                )
                # restrict processing to data that meets constraints for this evaluation frequency
                # add to class vars
                self.evalFreq.append(evalFreq[eIdx])
                # use process reduced - only the input channels from the remote reference
                numSolveWindows, obs, reg = self.prepareLinearEqn(
                    evalFreqData[eIdx, localWinIndices]
                )
                out, var = self.robustProcess(numSolveWindows, obs, reg)
                self.impedances.append(out)
                self.variances.append(var)

        if len(self.evalFreq) == 0:
            self.printWarning(
                "No data was found at any decimation level for insite {}, outsite {}, remotesite {} and specdir {}".format(
                    self.inSite, self.outSite, self.remoteSite, self.winSelector.specdir
                )
            )
            return

        # write out all the data
        self.writeTF(
            self.winSelector.specdir,
            self.postpend,
            self.evalFreq,
            self.impedances,
            self.variances,
            remotesite=self.remoteSite,
            remotechans=self.remoteChannels,
        )

    def checkRemote(self):
        check = True
        check = check and self.remoteSize == self.inSize
        check = check and self.remoteChannels == self.inChannels
        return check

    # PREPARE ROUTINES
    def prepareLinearEqn(self, data):
        # prepare observations and regressors for linear processing
        numWindows = data.shape[0]
        numWindows, data = self.checkForBadValues(numWindows, data)
        # for each output variable, have ninput regressor variables
        # let's construct our arrays
        obs = np.empty(
            shape=(self.outSize, self.remoteSize * numWindows), dtype="complex"
        )
        reg = np.empty(
            shape=(self.outSize, self.remoteSize * numWindows, self.inSize),
            dtype="complex",
        )
        for iW in range(0, numWindows):
            iOffset = iW * self.remoteSize
            for i in range(0, self.outSize):
                for j in range(0, self.remoteSize):
                    # this is the observation row where,i is the observed output
                    obs[i, iOffset + j] = data[iW, self.inSize + i, j]
                    for k in range(0, self.inSize):
                        reg[i, iOffset + j, k] = data[iW, k, j]
        return numWindows, obs, reg

    # SOLVER ROUTINES
    def robustProcess(self, numWindows, obs, reg):
        # do the chatterjeeMachlerMod robust processing for a single evaluation frequency
        # create array for output
        output = np.empty(shape=(self.outSize, self.inSize), dtype="complex")
        varOutput = np.empty(shape=(self.outSize, self.inSize), dtype="float")
        # solve
        for i in range(0, self.outSize):
            observation = obs[i, :]
            predictors = reg[i, :, :]
            # save the output
            out, resids, weights = chatterjeeMachler(
                predictors, observation, intercept=self.intercept
            )
            # out, resids, scale, weights = mmestimateModel(predictors, observation, intercept=self.intercept)

            # now take the weights, apply to the observations and predictors, stack the appropriate rows and test
            observation2 = np.zeros(shape=(self.remoteSize), dtype="complex")
            predictors2 = np.zeros(
                shape=(self.remoteSize, self.inSize), dtype="complex"
            )
            for iChan in range(0, self.remoteSize):
                # now need to have my indexing array
                indexArray = np.arange(
                    iChan, numWindows * self.remoteSize, self.remoteSize
                )
                weightsLim = weights[indexArray]
                # weightsLim = weightsLim/np.sum(weightsLim) # normalise weights to 1
                observation2[iChan] = (
                    np.sum(obs[i, indexArray] * weightsLim) / numWindows
                )
                # now for the regressors
                for j in range(0, self.inSize):
                    predictors2[iChan, j] = (
                        np.sum(reg[i, indexArray, j] * weightsLim) / numWindows
                    )
            out, resids, weights = chatterjeeMachler(
                predictors2, observation2, intercept=self.intercept
            )
            # out, resids, scale, weights = mmestimateModel(predictors2, observation2, intercept=self.intercept)

            # now calculate out the varainces - have the solution out, have the weights
            # recalculate out the residuals with the final solution
            # calculate standard deviation of residuals
            # and then use chatterjee machler formula to estimate variances
            # this needs work - better to use an empirical bootstrap method, but this will do for now
            resids = np.absolute(observation - np.dot(predictors, out))
            scale = sampleMAD0(
                resids
            )  # some measure of standard deviation, rather than using the standard deviation
            residsVar = scale * scale
            varPred = np.dot(hermitianTranspose(predictors), weights * predictors)
            varPred = np.linalg.inv(varPred)  # this is a pxp matrix
            varOut = 1.91472 * residsVar * varPred
            varOut = np.diag(varOut).real  # this should be a real number

            if self.intercept:
                output[i] = out[1:]
                varOutput[i] = varOut[1:]
            else:
                output[i] = out
                varOutput[i] = varOut

        return output, varOutput

    def olsProcess(self, numWindows, obs, reg):
        # ordinary least squares solution
        # create array for output
        output = np.empty(shape=(self.outSize, self.inSize), dtype="complex")
        # solve
        for i in range(0, self.outSize):
            observation = obs[i, :]
            predictors = reg[i, :, :]
            # save the output
            out, resids, squareResid, rank, s = olsModel(
                predictors, observation, intercept=self.intercept
            )
            if self.intercept:
                output[i] = out[1:]
            else:
                output[i] = out
        return output

    def stackedProcess(self, data):
        # then do various sums
        numWindows = data.shape[0]
        numWindows, data = self.checkForBadValues(numWindows, data)
        # unweighted sum (i.e. normal solution)
        unWeightedSum = np.sum(data, axis=0)
        unWeightedSum = unWeightedSum / numWindows
        # for each output variable, have ninput regressor variables
        # let's construct our arrays
        obs = np.empty(shape=(self.outSize, self.remoteSize), dtype="complex")
        reg = np.empty(
            shape=(self.outSize, self.remoteSize, self.inSize), dtype="complex"
        )
        for i in range(0, self.outSize):
            for j in range(0, self.remoteSize):
                obs[i, j] = unWeightedSum[self.inSize + i, j]
                for k in range(0, self.inSize):
                    reg[i, j, k] = unWeightedSum[k, j]
        # create array for output
        output = np.empty(shape=(self.outSize, self.inSize), dtype="complex")
        for i in range(0, self.outSize):
            observation = obs[i, :]
            predictors = reg[i, :, :]
            # save the output
            out, resids, scale, weights = mmestimateModel(
                predictors, observation, intercept=self.intercept
            )
            if self.intercept:
                output[i] = out[1:]
            else:
                output[i] = out
        return output

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        textLst = []
        textLst.append("In Site = {}".format(self.inSite))
        textLst.append("In Channels = {}".format(self.inChannels))
        textLst.append("Out Site = {}".format(self.outSite))
        textLst.append("Out Channels = {}".format(self.outChannels))
        textLst.append("Remote Site = {}".format(self.remoteSite))
        textLst.append("Remote Channels = {}".format(self.remoteChannels))
        return textLst




import numpy as np
import scipy.signal as signal
from datetime import datetime, timedelta
from typing import Dict

# import from package
from resistics.dataObjects.timeData import TimeData
from resistics.utilities.utilsPrint import generalPrint, arrayToStringInt
from resistics.utilities.utilsMath import intdiv


def normalise(timeData: TimeData, inplace: bool = True) -> TimeData:
    """Normalise time data
    
    Parameters
    ----------
    timeData : TimeData
        timeData to normalise
    inplace : bool, optional
        Whether to manipulate the data inplace

    Returns
    -------
    TimeData
        Normalised time data
    """

    if not inplace:
        timeData = timeData.copy()
    timeData.data = normaliseData(timeData.data)
    timeData.addComment("Data normalised")
    return timeData


def normaliseData(data: Dict[str, np.ndarray]) -> Dict[str, np.ndarray]:
    """Normalise array data
    
    Normalisation is done by dividing by the result of numpy.norm of the data

    Parameters
    ----------
    data : Dict
        Dictionary with channel as keys and data as values

    Returns
    -------
    Dict
        Dictionary with channel as keys and normalised data as values
    """

    for c in data:
        data[c] = data[c] / np.linalg.norm(data[c])
    return data


def lowPass(timeData: TimeData, cutoff: float, inplace: bool = True) -> TimeData:
    """Lowpass butterworth filter for time data
    
    Parameters
    ----------
    timeData : TimeData
        timeData to filter
    cutoff : float
        Cutoff frequency in Hz
    inplace : bool, optional
        Whether to manipulate the data inplace

    Returns
    -------
    TimeData
        Filtered time data
    """

    if not inplace:
        timeData = timeData.copy()
    timeData.data = lowPassData(timeData.data, timeData.sampleFreq, cutoff)
    timeData.addComment("Low pass filter applied with cutoff {} Hz".format(cutoff))
    return timeData


def lowPassData(
    data: Dict[str, np.ndarray], sampleFreq: float, cutoff: float, order: int = 5
) -> Dict[str, np.ndarray]:
    """Lowpass butterworth filter for array data
    
    Parameters
    ----------
    data : Dict
        Dictionary with channel as keys and data as values
    cutoff : float
        Cutoff frequency in Hz

    Returns
    -------
    data : Dict
        Dictionary with channel as keys and data as values
    """

    # create the filter
    normalisedCutoff = 2.0 * cutoff / sampleFreq
    b, a = signal.butter(order, normalisedCutoff, btype="lowpass", analog=False)
    # filter each channel
    return filterData(data, b, a)


def highPass(timeData: TimeData, cutoff: float, inplace: bool = True) -> TimeData:
    """Highpass butterworth filter for time data
    
    Parameters
    ----------
    timeData : TimeData
        timeData to filter
    cutoff : float
        Cutoff frequency in Hz
    inplace : bool, optional
        Whether to manipulate the data inplace        

    Returns
    -------
    TimeData
        Filtered time data
    """

    if not inplace:
        timeData = timeData.copy()
    timeData.data = highPassData(timeData.data, timeData.sampleFreq, cutoff)
    timeData.addComment("High pass filter applied with cutoff {} Hz".format(cutoff))
    return timeData


def highPassData(
    data: Dict[str, np.ndarray], sampleFreq: float, cutoff: float, order: int = 5
):
    """Highpass butterworth filter for array data
    
    Parameters
    ----------
    data : Dict
        Dictionary with channel as keys and data as values
    cutoff : float
        Cutoff frequency in Hz

    Returns
    -------
    data : Dict
        Dictionary with channel as keys and data as values
    """

    # create the filter
    normalisedCutoff = 2.0 * cutoff / sampleFreq
    b, a = signal.butter(order, normalisedCutoff, btype="highpass", analog=False)
    return filterData(data, b, a)


def bandPass(timeData: TimeData, cutoffLow: float, cutoffHigh: float, inplace: bool = True) -> TimeData:
    """Bandpass butterworth filter for time data
    
    Parameters
    ----------
    timeData : TimeData
        timeData to filter
    cutoff : float
        Cutoff frequency in Hz
    inplace : bool, optional
        Whether to manipulate the data inplace        

    Returns
    -------
    TimeData
        Filtered time data
    """

    if not inplace:
        timeData = timeData.copy()
    timeData.data = bandPassData(
        timeData.data, timeData.sampleFreq, cutoffLow, cutoffHigh
    )
    timeData.addComment(
        "Band pass filter applied with cutoffs {} Hz and {} Hz".format(
            cutoffLow, cutoffHigh
        )
    )
    return timeData


def bandPassData(
    data: Dict[str, np.ndarray],
    sampleFreq: float,
    cutoffLow: float,
    cutoffHigh: float,
    order: int = 5,
):
    """Bandpass butterworth filter for array data
    
    Parameters
    ----------
    data : Dict
        Dictionary with channel as keys and data as values
    cutoff : float
        Cutoff frequency in Hz

    Returns
    -------
    data : Dict
        Dictionary with channel as keys and data as values
    """

    # create the filter
    normalisedCutoffLow = 2.0 * cutoffLow / sampleFreq
    normalisedCutoffHigh = 2.0 * cutoffHigh / sampleFreq
    b, a = signal.butter(
        order,
        [normalisedCutoffLow, normalisedCutoffHigh],
        btype="bandpass",
        analog=False,
    )
    return filterData(data, b, a)


def filterData(data: Dict[str, np.ndarray], b, a) -> Dict[str, np.ndarray]:
    """Butterworth filter for array data
    
    Parameters
    ----------
    data : Dict
        Dictionary with channel as keys and data as values
    b : array_like
        The numerator coefficient vector of the filter
    a : array_like
        The denominator coefficient vector of the filter. If a[0] is not 1, then both a and b are normalized by a[0]

    Returns
    -------
    data : Dict
        Dictionary with channel as keys and data as values
    """

    filteredData = {}
    for c in data:
        filteredData[c] = signal.filtfilt(b, a, data[c], method="gust", irlen=500)
    return filteredData


def notchFilter(timeData: TimeData, notch: float, inplace: bool = True) -> TimeData:
    """Bandpass butterworth filter for time data
    
    Parameters
    ----------
    timeData : TimeData
        timeData to filter
    notch : float
        Frequency to notch filter in Hz
    inplace : bool, optional
        Whether to manipulate the data inplace        

    Returns
    -------
    TimeData
        Filtered time data
    """

    if not inplace:
        timeData = timeData.copy()
    timeData.data = notchFilterData(
        timeData.data, timeData.sampleFreq, notch, notch / 5.0
    )
    timeData.addComment("Notch filter applied at {} Hz".format(notch))
    return timeData


def notchFilterData(
    data: Dict[str, np.ndarray], sampleFreq: float, notch: float, band: float
) -> Dict[str, np.ndarray]:
    """Notch filter array data

    Parameters
    ----------
    data : Dict
        Dictionary with channel as keys and data as values
    sampleFreq : float
        Sampling frequency in Hz
    notch : float
        Frequency to notch in Hz
    band : float   
        The bandwidth around the centerline freqency that you wish to filter

    Returns
    -------
    data : Dict
        Dictionary with channel as keys and data as values    
    """

    # set parameters
    nyq = sampleFreq / 2.0
    low = notch - band / 2.0
    high = notch + band / 2.0
    low = low / nyq
    high = high / nyq
    # filter
    order = 2
    filter_type = "bessel"
    filteredData = {}
    for c in data:
        b, a = signal.iirfilter(
            order, [low, high], btype="bandstop", analog=False, ftype=filter_type
        )
        filteredData[c] = signal.lfilter(b, a, data[c])
    return filteredData


def resample(timeData: TimeData, resampFreq: float, inplace: bool = True) -> TimeData:
    """Resample time data
    
    Parameters
    ----------
    timeData : TimeData
        timeData to filter
    resampFreq : float
        The frequency to resample to
    inplace : bool, optional
        Whether to manipulate the data inplace        

    Returns
    -------
    TimeData
        Filtered time data
    """

    if not inplace:
        timeData = timeData.copy()
    timeData.data = resampleData(timeData.data, timeData.sampleFreq, resampFreq)
    # update the time info
    timeData.sampleFreq = resampFreq
    timeData.numSamples = timeData.data[timeData.chans[0]].size
    timeData.stopTime = timeData.startTime + timedelta(
        seconds=(1.0 / timeData.sampleFreq) * (timeData.numSamples - 1)
    )
    return timeData


def resampleData(
    data: Dict[str, np.ndarray], sampleFreq: float, sampleFreqNew: float
) -> Dict[str, np.ndarray]:
    """Resample array data
    
    Resample the data using the polyphase method which does not assume periodicity
    Calculate the upsample and then the downsampling rate and using polyphase filtering, the final sample rate is: 
    (up / down) * original sample rate
    Therefore, to get a sampling frequency of sampleFreqNew, want:
    (sampleFreqNew / sampleFreq) * sampleFreq
    Use the fractions library to get up and down as integers which they are required to be.

    Parameters
    ----------
    data : Dict
        Dictionary with channel as keys and data as values
    sampleFreq : float
        The current sampling frequency in Hz
    sampleFreqNew : float
        The sampling frequency in Hz to resample to

    Returns
    -------
    data : Dict
        Dictionary with channel as keys and data as values
    """

    from fractions import Fraction
    frac = Fraction(
        1.0 / sampleFreq
    ).limit_denominator()  # because this is most probably a float
    frac = Fraction(frac * int(sampleFreqNew))
    frac.limit_denominator()


    # otherwise, normal polyphase filtering
    resampleData = {}
    for c in data:
        resampleData[c] = signal.resample_poly(
            data[c], frac.numerator, frac.denominator
        )
    return resampleData


def downsampleData(data: Dict[str, np.ndarray], downsampleFactor: int) -> Dict[str, np.ndarray]:
    """Decimate array data
    
    Parameters
    ----------
    data : Dict
        Dictionary with channel as keys and data as values
    downsampleFactor : int
        The factor to downsample the data by

    Returns
    -------
    data : Dict
        Dictionary with channel as keys and data as values
    """
    
    # if downsampleFactor is 1, nothing to do
    if downsampleFactor == 1:
        return data

    # a downsample factor should not be greater than 13
    # hence factorise downsampleFactors that are greater than this
    if downsampleFactor > 13:
        downsamples = factorise(downsampleFactor)
        generalPrint(
            "Decimation",
            "Downsample factor {} greater than 13. Downsampling will be performed in multiple steps of {}".format(
                downsampleFactor, arrayToStringInt(downsamples)
            ),
        )
    else:
        downsamples = [downsampleFactor]

    # downsample for each factor in downsamples
    for factor in downsamples:
        for c in data:
            data[c] = signal.decimate(data[c], factor, zero_phase=True)
    return data


def factorise(number: int):
    """Factorise a number to avoid too large a downsample factor

    Parameters
    ----------
    number : int
        The number to factorise
    
    Returns
    -------
    List[int]
        The downsampling factors to use

    Notes
    -----
    There's a few pathological cases here that are being ignored. For example, what if the downsample factor is the product of two primes greater than 13.
    """
    import primefac
    factors = list(primefac.primefac(number))
    downsamples = []

    val = 1
    for f in factors:
        test = val * f
        if test > 13:
            downsamples.append(val)
            val = 1
        val = val * f
    # logic: on the last value of f, val*f is tested
    # if this is greater than 13, the previous val is added, which leaves one factor leftover
    # if not greater than 13, then this is not added either
    # so append the last value. the only situation in which this fails is the last factor itself is over 13.
    downsamples.append(val)
    return downsamples




import numpy as np
import math
from datetime import datetime, timedelta
from typing import List

# import from package
from resistics.calculators.calculator import Calculator
from resistics.dataObjects.timeData import TimeData
from resistics.utilities.utilsConfig import loadConfig
from resistics.utilities.utilsPrint import blockPrint


class Windower(Calculator):
    """Class for windowing time data with overlaps 

    Given time data for a decimation level and a reference time, the windower calculates the number of windows and returns them.

    Attributes
    ----------
    timeData : TimeData
        TimeData object to window
    numSamples : int
        The number of samples in each window
    sampleFreq : float
        Sampling frequency of data in Hz
    winSize : int
        Number of samples in a window
    winDuration : float
        Duration of a window in seconds
    winOlap : np.ndarray
        Number of overlap samples
    chans : List[str]
        List of chans in timeData
    refTime : datetime
        Project reference time
    dataTime : datetime
        Time of timeData first sample
    minWindows : int
        Minimum number of windows required
    winOffset : int
        Offset between local and global window indexing
    firstWindowTime: datetime
        Time of the first window in the TimeData
    winSamples : np.ndarray[int]
        Two dimensional, the sample start and end for each window
    winTimes : 
    winActive : np.ndarray[bool]
        Boolean flagging whether window is active or not

    Methods
    -------
    __init__(refTime, timeData, winSize, winOlap)
        Initialise windower
    getWindowActive(iW)

    getGlobalIndex(iW)
        Return the global index (relative to reference time) given local window index
    getData(iWindow)
        Get timeData for local window iWindow
    getDataGlobal(iGlobal)
        Get timeData for glonal index iGlobal
    initialiseWindows()
        Calculate number of windows, local to global index mapping and local window offset from global
    printList()
        Class status returned as list of strings  
    printWindowTimes()
        Print window times  
    printWindowTimeList()
        Window time information returned as List of strings               
    """

    def __init__(
        self, refTime: datetime, timeData: TimeData, winSize: int, winOlap: int
    ):
        self.timeData: TimeData = timeData
        self.numSamples: int = timeData.numSamples
        self.sampleFreq: float = timeData.sampleFreq
        self.winSize: int = winSize
        self.winDuration: float = (winSize - 1) / timeData.sampleFreq
        self.winOlap: int = winOlap
        self.chans: List[str] = timeData.chans
        # refTime and dataTime are already datetime objects
        self.refTime: datetime = refTime
        self.dataTime: datetime = timeData.startTime
        # min window warning setting
        config = loadConfig()
        self.minWindows: int = config["Window"]["minwindows"]
        # initialise
        self.initialiseWindows()
        self.calcWindowTimes()

    def getWindowActive(self, iWindow: int) -> bool:
        """Returns active status of local index

        Parameters
        ----------
        iWindow : int
            Local index of window

        Returns
        -------
        active : bool
            True if active, false if not
        """

        return self.winActive[iWindow]

    def getGlobalIndex(self, iWindow: int) -> int:
        """Returns global index for local index

        Parameters
        ----------
        iWindow : int
            Local index of window

        Returns
        -------
        iGlobal : int
            Global index
        """

        return iWindow + self.winOffset

    def getData(self, iWindow: int) -> TimeData:
        """Returns time window data for local index

        Parameters
        ----------
        iWindow : int
            Local index of window

        Returns
        -------
        windowData : TimeData
            TimeData object with the window data
        """

        winSamples = self.winSamples[iWindow]
        winData = {}
        for c in self.chans:
            winData[c] = self.timeData.data[c][
                winSamples[0] : winSamples[1] + 1
            ]  # add 1 because numpy indexing like this is not inclusive
        globalWindow = self.winTimes[iWindow][0]
        winStartTime = self.winTimes[iWindow][1]
        winStopTime = self.winTimes[iWindow][2]
        return TimeData(
            sampleFreq=self.sampleFreq,
            startTime=winStartTime,
            stopTime=winStopTime,
            data=winData,
            comments=self.timeData.comments
            + ["Local window iW, global window {}".format(globalWindow)],
        )

    def getDataGlobal(self, iGlobal: int) -> TimeData:
        """Returns time window data for global index

        Parameters
        ----------
        iGlobal : int
            Global index of window

        Returns
        -------
        windowData : TimeData
            TimeData object with the window data
        """

        iWindow = iGlobal - self.winOffset
        return self.getData(iWindow)

    def initialiseWindows(self):
        """Calculate all the window information

        For timeData and referenceTime, initialiseWindows calculates the number of windows (after the reference time) and the global indices of the windows relative to the reference time
        Stores the offset between local indices and the global indices
        """

        # have a reference time
        # the first window starts there
        deltaRefStart = self.dataTime - self.refTime
        if deltaRefStart.total_seconds() < 0:
            self.printWarning(
                "Reference time is after start of recording. Stuff may go wrong!"
            )
        # increment of window start times
        # -1 because inclusive of sample at start
        winStartIncrement = 1.0 * (self.winSize - self.winOlap) / self.sampleFreq
        # calculate number of windows started before reference time
        # and then by taking the ceiling, find the global index of the first window in the data
        self.winOffset = int(
            math.ceil(deltaRefStart.total_seconds() / winStartIncrement)
        )
        # calculate start time of first global window
        offsetSeconds = self.winOffset * winStartIncrement
        # calculate the first window time
        self.firstWindowTime = self.refTime + timedelta(seconds=offsetSeconds)
        # calculate first sample
        deltaStart = self.firstWindowTime - self.dataTime
        sampleStart = deltaStart.total_seconds() * self.sampleFreq
        # next calculate number of windows
        # sample start is the first sample
        # window size is window size inclusive of first sample
        winStart = sampleStart
        winEnd = sampleStart + self.winSize - 1
        winStartOff = self.winSize - self.winOlap
        winSamples = []
        while winEnd < self.numSamples:
            winSamples.append([winStart, winEnd])
            winStart = winStart + winStartOff
            winEnd = winStart + self.winSize - 1
        self.numWindows = len(winSamples)
        # warn if number of windows is small
        if self.numWindows < self.minWindows:
            self.printWarning(
                "Number of windows in data is small - consider stopping decimation"
            )
        # save winSamples as numpy list in class
        self.winSamples = np.array(winSamples, dtype=int)
        # set all windows initially to active
        self.winActive = np.ones(shape=(self.numWindows), dtype=bool)

    def calcWindowTimes(self) -> None:
        """Calculate start and stop times for each window"""

        self.winTimes = []
        iW = 0
        for samples in self.winSamples:
            start = samples[0]
            stop = samples[1]
            win = []
            # global index
            win.append(self.winOffset + iW)
            # start time
            deltaStart = timedelta(seconds=start / self.sampleFreq)
            timeStart = self.dataTime + deltaStart
            deltaEnd = timedelta(seconds=stop / self.sampleFreq)
            timeEnd = self.dataTime + deltaEnd
            # samples2end = self.winSize - 1 # need to remove initial sample
            # timeEnd = timeStart + timedelta(seconds=samples2end/self.sampleFreq)
            win.append(timeStart)
            win.append(timeEnd)
            self.winTimes.append(win)
            iW = iW + 1

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        textLst = []
        textLst.append("Sample freq. [Hz] = {:f}".format(self.sampleFreq))
        textLst.append("Window size = {:d}".format(self.winSize))
        textLst.append("Overlap size = {:d}".format(self.winOlap))
        textLst.append("Window duration [s] = {:.3f}".format(self.winDuration))
        textLst.append("Reference time {}".format(self.refTime))
        textLst.append("Data start time {}".format(self.dataTime))
        textLst.append(
            "Number of complete windows in data = {:d}".format(self.numWindows)
        )
        if self.numWindows < self.minWindows:
            textLst.append(
                "Number of windows in data is small - consider stopping decimation"
            )
        if self.numWindows > 0:
            textLst.append(
                "Global index of first window from reference time = {}".format(
                    self.winOffset
                )
            )
            textLst.append(
                "First window starts at time {}, sample {:d}".format(
                    self.firstWindowTime, self.winSamples[0, 0]
                )
            )
        return textLst

    def printWindowTimes(self) -> None:
        """Print the times of the windows"""

        blockPrint("Windower::window times", self.printWindowTimeList())

    def printWindowTimeList(self) -> List:
        """Window time information as a list of strings

        Returns
        -------
        out : list
            List of strings with information
        """

        winTimes = self.winTimes
        winSamples = self.winSamples
        textLst = []
        textLst.append(
            "NOTE: Sample ranges are inclusive, to get number of samples, use: sample end - sample start + 1"
        )
        for win, winS in zip(winTimes, winSamples):
            textLst.append(
                "Global index = {:d}, start time = {}, end time = {}, start sample = {:d}, end sample = {:d}".format(
                    win[0], win[1], win[2], winS[0], winS[1]
                )
            )
        return textLst




import os
import numpy as np
import math
from datetime import datetime, timedelta
from typing import List, Tuple

# import from package
from resistics.ioHandlers.ioHandler import IOHandler
from resistics.dataObjects.calibrationData import CalibrationData
from resistics.utilities.utilsIO import checkAndMakeDir
from resistics.utilities.utilsPrint import listToString


class CalibrationIO(IOHandler):
    """Class for reading and writing calibration files

    Calibration files hold impulse response data for instruments and sensors that needs to be removed from time series data.

    Calibration data is stored in:
    project -> calData

    Attributes
    ----------
    datafile : str
        Path to calibration data file
    fileformat : str
        The file format of the calibration file
    chopper : bool
        Boolean flag for reading chopper on or off or writing chopper on or off
    extend : bool
        Boolean flag for extending (extrapolating) the calibration data

    Methods
    -------
    __init__(datapath, fileformat, chopper, extend)
        Initialise the CalibrationIO object.
    refresh(datapath, fileformat, chopper, extend)
        Refresh the CalibrationIO parameters
    read()
        Read calibration file
    readInductionData()
        Read internal format calibration file for induction coils        
    readMetronixData()
        Read metronix style calibration file for induction coils
    readRSPData()
        Read RSP style calibration file for induction coils
    readRSPXData()
        Read RSPX style calibration file for induction coils
    readInternalFormat()
        Read internal format calibration file
    unitCalibration()
        A unit calibration, used if nothing else is found
    sortCalData(data)
        Sort calibration data into ascending order (lower to higher)
    extendCalData(data)
        Extrapolate the calibration data to higher and lower frequencies
    linesToArray(dataLines)
        Convert a set of data lines in a calibration file into a list    
    writeInternalFormat()
        Write internal format calibration file
    writeInternalTemplate()
        Write a template file for internal calibration format
    printList()
        Class status returned as list of strings       
	"""

    def __init__(
        self,
        datafile: str = "",
        fileformat: str = "induction",
        chopper: bool = False,
        extend: bool = True,
    ) -> None:
        """Initialise CalibrationIO

        Parameters
        ----------
        datafile : str, optional, optional
            Path to calibration data file
        format : str, optional, optional
            Calibration file format
        chopper : bool, optional
            Boolean flag for reading chopper on or off data, mostly used with metronix data. Default is False.
        extend : bool, optional
            Boolean flag for extrapolating the calibration data. Default is True.
        """

        self.refresh(datafile, fileformat, chopper, extend)

    def refresh(
        self,
        datafile: str = "",
        fileformat: str = "induction",
        chopper: bool = False,
        extend: bool = True,
    ) -> None:
        """Refresh the CalibrationIO parameters

        Parameters
        ----------
        datafile : str, optional
            Path to calibration data file
        format : str, optional
            Calibration file format
        chopper : bool
            Boolean flag for reading chopper on or off data, mostly used with metronix data
        extend : bool
            Boolean flag for extrapolating the calibration data
        """

        self.datafile: str = datafile
        self.fileformat: str = fileformat
        self.chopper: bool = chopper
        self.extend: bool = extend

    def read(self) -> CalibrationData:
        """Read data from calibration file and return CalibrationData object

        Notes
        -----
        Data for magnetic channel calibration is returned with units: F [Hz], Magnitude [mV/nT], Phase [radians]

        Returns
        -------
        CalibrationData
            A calibration data object
        """

        if self.fileformat == "metronix":
            return self.readMetronixData()
        elif self.fileformat == "rsp":
            return self.readRSPData()
        elif self.fileformat == "rspx":
            return self.readRSPXData()
        elif self.fileformat == "induction":
            return self.readInductionData()
        else:
            return self.unitCalibration()

    def readInductionData(self):
        """Read data from a text based calibration file for induction coils
        
        Text based calibration files should be used when reading of other formats appears to be failing or are not supported. 

        Notes
        -----
        Data in file is assumed to be in units: F [Hz], Magnitude [mv/nT], Phase [degrees or radians]
        Data is returned with units: F [Hz], Magnitude [mV/nT], Phase [radians]

        Returns
        -------
        CalibrationData
            A calibration data object
        """

        calData = self.readInternalFormat()
        return calData

    def readMetronixData(self) -> CalibrationData:
        """Read data from metronix calibration file
        
        Notes
        -----
        Metronix data is in units: F [Hz], Magnitude [V/nT*Hz], Phase [deg] for both chopper on and off.
        Data is returned with units: F [Hz], Magnitude [mV/nT], Phase [radians].  

        Returns
        -------
        CalibrationData
            A calibration data object
        """

        with open(self.datafile, "r") as f:
            lines = f.readlines()
        numLines = len(lines)

        # serial and sensor - there is no sensor information in the file
        serial: int = 1
        sensor: str = ""
        # no static gain - already included
        staticGain: float = 1
        # variables to save line numbers
        chopperOn: int = 0
        chopperOff: int = 0
        # find locations for chopperOn and chopperOff
        for il in range(0, numLines):
            # remove whitespace and new line characters
            lines[il] = lines[il].strip()
            if "Magnetometer" in lines[il]:
                split1 = lines[il].split(":")[1].strip()
                serial = int(split1.split()[0].strip())
            if "Chopper On" in lines[il]:
                chopperOn = il
            if "Chopper Off" in lines[il]:
                chopperOff = il

        # get the part of the file required depending on chopper on or off
        dataLines: List = []
        dataReadFrom: int = chopperOff
        if self.chopper:
            dataReadFrom = chopperOn
        # get the data - starting from the next line
        il = dataReadFrom + 1
        while il < numLines and lines[il] != "":
            # save line then increment
            dataLines.append(lines[il])
            il = il + 1

        # get the data as an array
        data = self.linesToArray(dataLines)
        # sort and extend
        data = self.sortCalData(data)
        if self.extend:
            data = self.extendCalData(data)
        # unit manipulation - change V/(nT*Hz) to mV/nT
        data[:, 1] = data[:, 1] * data[:, 0] * 1000
        # unit manipulation - change phase to radians
        data[:, 2] = data[:, 2] * (math.pi / 180)

        return CalibrationData(
            self.datafile,
            data[:, 0],
            data[:, 1],
            data[:, 2],
            staticGain=staticGain,
            chopper=self.chopper,
            serial=serial,
            sensor=sensor,
        )

    def readRSPData(self) -> CalibrationData:
        """Read data from a RSP calibration file
        
        Notes
        -----        
        RSP data is in units: F [Hz], Magnitude [mv/nT], Phase [deg]
        Data is returned with units: F [Hz], Magnitude [mV/nT], Phase [radians]

        Returns
        -------
        CalibrationData
            A calibration data object
        """

        with open(self.datafile, "r") as f:
            lines = f.readlines()
        numLines = len(lines)

        serial: int = 1
        staticGain: float = 1
        sensor: str = ""
        dataReadFrom: int = 0
        for il in range(0, numLines):
            # remove whitespace and new line characters
            lines[il] = lines[il].strip()
            # find serial
            if "induction coil no" in lines[il]:
                split1 = lines[il].split(":")[1]
                serial = int(split1.split("-")[0].strip())
            # find sensor
            if "SensorType" in lines[il]:
                sensor = lines[il].split()[1]
            # find static gain value
            if "StaticGain" in lines[il]:
                staticGain = float(lines[il].split()[1])
            if "FREQUENCY" in lines[il]:
                dataReadFrom = il
        dataLines = []
        il = dataReadFrom + 2
        # get part of file desired
        while il < numLines and lines[il] != "":
            # save line then increment
            dataLines.append(lines[il])
            il = il + 1

        # get the data as an array
        data = self.linesToArray(dataLines)
        # unit manipulation - change phase to radians and apply static gain
        data[:, 1] = data[:, 1] * staticGain
        data[:, 2] = data[:, 2] * (math.pi / 180)
        # sort and extend
        data = self.sortCalData(data)
        if self.extend:
            data = self.extendCalData(data)

        return CalibrationData(
            self.datafile,
            data[:, 0],
            data[:, 1],
            data[:, 2],
            staticGain,
            chopper=self.chopper,
            sensor=sensor,
            serial=serial,
        )

    def readRSPXData(self) -> CalibrationData:
        """Read data from calibration file
        
        Notes
        -----
        RSPX data is in units: F [Hz], Magnitude [mv/nT], Phase [deg]
        Data is returned with units: F [Hz], Magnitude [mV/nT], Phase [radians]

        Returns
        -------
        CalibrationData
            A calibration data object
        """

        # this is xml format - use EL tree
        import xml.etree.ElementTree as ET

        tree = ET.parse(self.datafile)
        root = tree.getroot()
        # serial
        serial: int = 1
        if root.find("SensorId") is not None:
            serial = int(root.find("SensorId").text)
        # sensor
        sensor: str = ""
        if root.find("SensorSpecification") is not None:
            sensor = root.find("SensorSpecification").text
        # static gain
        staticGain: float = 1.0
        if root.find("StaticGain") is not None:
            staticGain = float(root.find("StaticGain").text)
        # get the calibration data
        dataList = []
        for resp in root.findall("ResponseData"):
            dataList.append(
                [
                    float(resp.get("Frequency")),
                    float(resp.get("Magnitude")),
                    float(resp.get("Phase")),
                ]
            )
        # now create array
        data = np.array(dataList)
        # change phase to radians and apply static gain
        data[:, 1] = data[:, 1] * staticGain
        data[:, 2] = data[:, 2] * (math.pi / 180)
        # sort and extend
        data = self.sortCalData(data)
        if self.extend:
            data = self.extendCalData(data)

        return CalibrationData(
            self.datafile,
            data[:, 0],
            data[:, 1],
            data[:, 2],
            staticGain,
            chopper=self.chopper,
            sensor=sensor,
            serial=serial,
        )

    def readInternalFormat(self):
        """Read data from a text based calibration file for any type of instrument 

        Notes
        -----
        Text based calibration files should be used when reading of other formats appears to be failing or are not supported. Nothing is assumed or promised about the data units.

        Returns
        -------
        CalibrationData
            A calibration data object
        """

        with open(self.datafile, "r") as f:
            lines = f.readlines()

        for il, line in enumerate(lines):
            if "Serial" in line:
                serial = int(line.split("=")[1].strip())
            if "Sensor" in line:
                sensor = line.split("=")[1].strip()
            if "Static gain" in line:
                staticGain = float(line.split("=")[1].strip())
            if "Magnitude unit" in line:
                magnitudeUnit = line.split("=")[1].strip()
            if "Phase unit" in line:
                phaseUnit = line.split("=")[1].strip()
            if "Chopper" in line:
                chopperStr = line.split("=")[1].strip()
                if chopperStr == "True":
                    chopper = True
                else:
                    chopper = False
            if "CALIBRATION DATA" in line:
                dataReadFrom = il + 1
                break

        dataLines = lines[dataReadFrom:]
        data = self.linesToArray(dataLines)
        # sort and extend
        data = self.sortCalData(data)
        if self.extend:
            data = self.extendCalData(data)
        # apply static gain
        data[:, 1] = data[:, 1] * staticGain
        if phaseUnit == "degrees":
            data[:, 2] = data[:, 2] * (math.pi / 180)

        calData = CalibrationData(
            self.datafile,
            data[:, 0],
            data[:, 1],
            data[:, 2],
            staticGain,
            chopper=chopper,
            serial=serial,
            sensor=sensor,
        )
        calData.magnitudeUnit = magnitudeUnit
        calData.phaseUnit = "radians"
        return calData

    def unitCalibration(self) -> CalibrationData:
        """Return a unit calibration

        Returns
        -------
        CalibrationData
            A calibration data object
        """

        data = np.array([[0.0000001, 1, 0], [100000000, 1, 0]])
        staticGain = 1
        return CalibrationData(
            self.datafile, data[:, 0], data[:, 1], data[:, 2], staticGain
        )

    def sortCalData(self, data: np.ndarray) -> np.ndarray:
        """Sort calibration data by frequency ascending (low to high)

        Parameters
        ----------
        data : np.ndarray
            Unsorted calibration data   

        Returns
        -------
        data : np.ndarray
            Sorted calibration data
        """

        return data[data[:, 0].argsort()]

    def extendCalData(self, data: np.ndarray) -> np.ndarray:
        """Extend calibration data by frequency

        Add extra points at the start and end of the calibration data to ensure complete coverage with the time data. 

        Notes
        -----
        It is assumed that the calibration data is already sorted in ascending order (using sortCalData)

        Parameters
        ----------
        data : np.ndarray
            Calibration data   

        Returns
        -------
        data : np.ndarray
            Extended calibration data
        """

        if data[0, 0] > 0.0000001:
            # add a line at the top (low frequency) extending the calibration information
            data = np.vstack((np.array([0.0000001, data[0, 1], data[0, 2]]), data))
        if data[-1, 0] < 100000000:
            # add a line at the top (high frequency) extending the calibration information
            data = np.vstack((data, np.array([100000000, data[-1, 1], data[-1, 2]])))
        return data

    def linesToArray(self, dataLines: List) -> np.ndarray:
        """Convert data lines from a file to an array

        Parameters
        ----------
        dataLines : list
            Data lines read in from a file   

        Returns
        -------
        data : np.ndarray
            Data lines converted to a float array
        """

        # data to columns
        numData = len(dataLines)
        for il in range(0, numData):
            dataLines[il] = dataLines[il].split()
        return np.array(dataLines, dtype=float)

    def writeInternalFormat(self, calData: CalibrationData, filepath: str) -> None:
        """Write out a calibration data file

        Parameters
        ----------
        calibrationData : CalibrationData
            Calibration data to write out
        filepath : str
            The file to write out to
        """

        with open(filepath, "w") as f:
            f.write("Serial = {}\n".format(calData.serial))
            f.write("Sensor = {}\n".format(calData.sensor))
            f.write("Static gain = {}\n".format(calData.staticGain))
            f.write("Magnitude unit = {}\n".format(calData.magnitudeUnit))
            f.write("Phase unit = {}\n".format(calData.phaseUnit))
            f.write("Chopper = {}\n".format(calData.chopper))
            f.write("\n")
            f.write("CALIBRATION DATA\n")
            # remove static gain from magnitude
            magnitude = np.array(calData.magnitude)
            magnitude = magnitude / calData.staticGain
            # write out magnitude without static gain
            for ii in range(0, calData.numSamples):
                f.write(
                    "{:+.4e}  {:+.4e}  {:+.4e}\n".format(
                        calData.freqs[ii], magnitude[ii], calData.phase[ii]
                    )
                )

    def writeInternalTemplate(
        self,
        filepath: str,
        serial: int = 1,
        sensor: str = "",
        staticgain: float = 1.0,
        chopper: bool = False,
        magnitudeUnit: str = "mV/nT",
        phaseUnit: str = "degrees",
    ) -> None:
        """Write out a calibration data template file

        This make it easier for the creation of a custom calibration file

        Parameters
        ----------
        filepath : str
            The file to write out to
        serial : int, optional
            The serial number of the sensor/instrument
        sensor : str, optional
            The sensor type
        staticgain : float, optional
            The static gain
        chopper : bool, optional
            The chopper on or off
        magnitudeUnit : str, optional
            The units of the magnitude. This should be mV/nT because that is what the internal format expects
        phaseUnit : str, optional
            Can be either degrees or radians
        """

        with open(filepath, "w") as f:
            f.write("Serial = {}\n".format(serial))
            f.write("Sensor = {}\n".format(sensor))
            f.write("Static gain = {}\n".format(staticgain))
            f.write("Magnitude unit = {}\n".format(magnitudeUnit))
            f.write("Phase unit = {}\n".format(phaseUnit))
            f.write("Chopper = {}\n".format(chopper))
            f.write("\n")
            f.write("CALIBRATION DATA\n")                

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        if self.datafile == "":
            textLst.append(
                "No datafile given. Please set the datafile attribute of the class"
            )
        else:
            textLst.append("datafile = {}".format(self.datafile))
        textLst.append("Extend = {}".format(self.extend))
        return textLst



import os
import glob
from datetime import datetime, timedelta
import numpy as np
from typing import List, Dict, Union, Any

# import from package
from resistics.dataObjects.timeData import TimeData
from resistics.ioHandlers.ioHandler import IOHandler
from resistics.ioHandlers.dataReader import DataReader
from resistics.utilities.utilsIO import checkAndMakeDir


class DataWriter(IOHandler):
    """Base class for data writers

    All input readers provide headers in a common format

	Attributes
	----------
    outPath : str
        The path to write to
    extension: str
        The extension to give to the data files
    dtype : data type
        Data format to write out, default is np.int32
    headers : 
        Header information to write out
    chans : List[str]
        Channels to write out
    chanMap : Dict
        Map between channel and index in channel headers
    chanHeaders : List
        Channel specific headers

	Methods
	-------
	__init__()
		Initialise writer
    getOutPath()
        Get the path to write out to
    setOutPath(path)
        Set the path to write out to
    setExtension()
        For subclasses to set their extension
    setGlobalHeadersFromKeywords(headers, keywords)
        Set the global headers 
    setChanHeadersFromKeywords(chanHeaders, keywords)
        Set channel headers
    calcStopDateTime(sampleFreq, numSamples, datetimeStart)
        Calculate time series stop time
    globalHeaderwords()
        Get a list of the global headers of interest
    chanHeaderwords()
        Get a list of the channel headers of interest
    writeTemplateHeaderFiles(chans, chanFileMap, sampleFreq, numSamples, startDate)    
        Write out a set of template header files for a case in which ASCII data without headers is to be loaded in
    writeDataset(reader, **kwargs)
        Write an existing dataset as a different format
    writeData(headers, chanHeaders, timeData, **kwargs)
        Write data based on headers, channel headers and time series data
    write(headers, chanHeaders, chanMap, timeData, **kwargs)
        Write out a dataset
    writeHeaders(headers, chans, chanMap, chanHeaders)
        Write out headers
    writeComments(comments)
        Write out comments
    writeDataFiles(chans, timeData)
        Write out time series data - not implemented in base class
    printList()
        Class status returned as list of strings   
	"""

    def __init__(self):
        self.outPath: str = ""
        # in subclasses, extension might change i.e. .ascii
        self.setExtension()
        # data type - the format of the data being written out
        self.dtype = np.int32
        # information about data being written
        self.headers: Union[Dict, None] = None
        self.chans: Union[List[str], None] = None
        self.chanMap: Union[Dict[str, int], None] = None
        self.chanHeaders: Union[List, None] = None

    def getOutPath(self) -> str:
        """Get the out path

        Parameters
        ----------
        str
            The outpath defining where data is written
        """

        return self.outPath

    def setOutPath(self, path: str) -> None:
        """Set the out path

        Parameters
        ----------
        path : str
            The new outpath defining where data is written
        """

        self.outPath = path

    def setExtension(self) -> None:
        """For subclasses to set their own extension type"""
        
        self.extension = ".dat"

    def setGlobalHeadersFromKeywords(self, headers: Dict, keywords: Dict) -> Dict:
        """Set the global headers

        Before writing out data, global headers are set. The priority order is:

        1. keywords[headername] if headername exists in keywords
        2. headers[headername] if headername exists in headers
        3. "" where the header is not defined in either keywords or headers
        
        The reason keywords takes top priority is there may be instances where the headers defined in a reader may need to be altered due to processing of time data.

        Parameters
        ----------
        headers : Dict
            Dictionary of header values
        keywords : Dict
            A dictionary of header values to overwrite those in headers
        """

        globalHeaderwords = self.globalHeaderwords()
        for gH in globalHeaderwords:
            hdrVal = ""
            if gH in headers:
                hdrVal = headers[gH]
            if gH in keywords:
                hdrVal = keywords[gH]
            headers[gH] = hdrVal
        return headers

    def setChanHeadersFromKeywords(self, chanHeaders: List, keywords: Dict) -> List:
        """Set the channel headers

        Before writing out data, channel headers are set. The priority order is:

        1. keywords[headername] if headername exists in keywords
        2. headers[headername] if headername exists in headers
        3. "" where the channel header is not defined in either keywords or headers
        
        The reason keywords takes top priority is there may be instances where the headers defined in a reader may need to be altered due to processing of time data.

        Parameters
        ----------
        chanHeaders : List
            List of channel headers
        keywords : Dict
            A dictionary of header values to overwrite those in channel headers
        """

        chanHeaderwords = self.chanHeaderwords()
        for iChan in range(0, len(chanHeaders)):
            for cH in chanHeaderwords:
                hdrVal = ""
                if cH in chanHeaders[iChan]:
                    hdrVal = chanHeaders[iChan][cH]
                if cH in keywords:
                    hdrVal = keywords[cH]
                chanHeaders[iChan][cH] = hdrVal
        return chanHeaders

    def calcStopDateTime(
        self, sampleFreq: float, numSamples: int, datetimeStart: datetime
    ) -> datetime:
        """Calculate time of last sample 
        
        Parameters
        ----------
        sampleFreq : float
            Sampling frequency in Hz of the time series data
        numSamples : int
            The number of samples in the time series data
        datetimeStart : datetime
            The time of the first sample
        """

        # calculate duration in seconds
        # numSamples - 1 because have to remove the initial sample which is taken at start time
        duration = 1.0 * (numSamples - 1) / sampleFreq
        datetimeStop = datetimeStart + timedelta(seconds=duration)
        return datetimeStop

    def globalHeaderwords(self) -> List[str]:
        """Get a list of global headerwords to write out

        Returns
        -------
        List[str]
            A list of the global header words of interest for writing out
        """

        gHeaders = [
            "sample_freq",
            "num_samples",
            "start_time",
            "start_date",
            "stop_time",
            "stop_date",
            "meas_channels",
        ]
        return gHeaders

    def chanHeaderwords(self) -> List[str]:
        """Get a list of channel headerwords to write out

        Returns
        -------
        List[str]
            A list of the global header words of interest for writing out
        """

        cHeaders = [
            "sample_freq",
            "num_samples",
            "start_time",
            "start_date",
            "stop_time",
            "stop_date",
            "ats_data_file",
            "sensor_type",
            "channel_type",
            "ts_lsb",
            "scaling_applied",
            "pos_x1",
            "pos_x2",
            "pos_y1",
            "pos_y2",
            "pos_z1",
            "pos_z2",
            "sensor_sernum",
            "gain_stage1",
            "gain_stage2",
            "hchopper",
            "echopper",
        ]
        return cHeaders

    def writeTemplateHeaderFiles(
        self,
        chans: List[str],
        chanFileMap: Dict,
        sampleFreq: float,
        numSamples: int,
        startDate: str,
    ):
        """Write a set of blank headers

        Blank headers might be useful for reading in ascii files where no headers are existing. By giving a few header words, many options can be set 

        Parameters
        ----------
        chans : List[str]
            List of chans (e.g. Ex, Ey, Hx, Hy, Hz)
        chanFileMap : Dict[str, str]
            Map from channel to file 
        numsamples : int
            Number of samples of data
        startdate : str
            The start date of the recording in format %Y-%m-%d %H:%M:%S
        """

        # calculate start and end datetime
        datetimeStart = datetime.strptime(startDate, "%Y-%m-%d %H:%M:%S")
        datetimeStop = self.calcStopDateTime(sampleFreq, numSamples, datetimeStart)
        # set global header words
        globalKeywords = dict()
        globalKeywords["sample_freq"] = sampleFreq
        globalKeywords["num_samples"] = numSamples
        globalKeywords["start_date"] = datetimeStart.strftime("%Y-%m-%d")
        globalKeywords["start_time"] = datetimeStart.strftime("%H:%M:%S.%f")
        globalKeywords["stop_date"] = datetimeStop.strftime("%Y-%m-%d")
        globalKeywords["stop_time"] = datetimeStop.strftime("%H:%M:%S.%f")
        globalKeywords["meas_channels"] = len(chans)

        # empty dictionary so that some headers get defaulted
        emptyDict = dict()
        # set global headers for keyword arguments
        headers = self.setGlobalHeadersFromKeywords(emptyDict, globalKeywords)
        # set channel headers for keyword arguments
        chanMap: Dict = dict()
        chanHeaders: List[Dict] = list()
        for idx, chan in enumerate(chans):
            chanMap[chan] = idx
            chanKeywords = dict(globalKeywords)
            chanKeywords["scaling_applied"] = True
            chanKeywords["ts_lsb"] = 1
            chanKeywords["gain_stage1"] = 1
            chanKeywords["gain_stage2"] = 1
            chanKeywords["hchopper"] = 0
            chanKeywords["echopper"] = 0 
            chanKeywords["pos_x1"] = 0
            chanKeywords["pos_x2"] = 1
            chanKeywords["pos_y1"] = 0
            chanKeywords["pos_y2"] = 1
            chanKeywords["pos_z1"] = 0
            chanKeywords["pos_z2"] = 1  
            chanKeywords["sensor_sernum"] = 1                     
            chanHeaders.append(chanKeywords)
        # set the chan header words
        self.setChanHeadersFromKeywords(chanHeaders, emptyDict)
        for idx, chan in enumerate(chans):   
            # amend the data file in the chan headers     
            chanHeaders[idx]["ats_data_file"] = chanFileMap[chan]
            chanHeaders[idx]["channel_type"] = chan
            for cH in chanHeaders[idx].keys():
                if chanHeaders[idx][cH] == "":
                    chanHeaders[idx][cH] = "None"
        self.writeHeaders(headers, chans, chanMap, chanHeaders, rename=False)

    def writeDataset(self, reader: DataReader, physical: bool = True, **kwargs) -> None:
        """Write out a dataset by passing a data reader

        This method is intended to transform an existing dataset into internal format

        Parameters
        ----------
        reader : DataReader
            A list of the global header words of interest for writing out
        physical : bool, optional
            An optional flag designating whether to use physical samples or not. Default is true
        """

        if self.getOutPath() == "":
            self.printError("No output filepath given", quitRun=True)
        checkAndMakeDir(self.getOutPath())
        # write using information from a reader file
        headers = reader.getHeaders()
        chanHeaders, chanMap = reader.getChanHeaders()
        # now write depending on whether scaling_applied or not
        if physical:
            # make sure dataset is written out in float and with scaling_applied header set to True
            self.dtype = np.float32
            kwargs["scaling_applied"] = True
            # write out
            self.write(
                headers, chanHeaders, chanMap, reader.getPhysicalSamples(), **kwargs
            )
        else:
            # write out unscaled samples
            self.printWarning(
                "Wrinting out of unscaled samples is not recommended due to scaling differences between the formats."
            )
            self.printWarning(
                "Dataset will be written out but problems may be encountered in the future."
            )
            self.write(
                headers, chanHeaders, chanMap, reader.getUnscaledSamples(), **kwargs
            )

    def writeData(self, headers, chanHeaders, timeData, physical: bool = True, **kwargs):
        """Write out time data 

        This method requires the user to pass global headers and chan headers explicitly.

        Parameters
        ----------
        headers : Dict
            Dictionary of headers
        chanHeaders : List
            List of channel headers
        timeData : TimeData
            Time series data to write out
        physical : bool, optional
            An optional flag designating whether the data is in field units (i.e. all scalings have been applied). This will result in the scaling_applied header being set to True. Default value for physical is True (i.e. data is assumed to be in field units).
        """

        if self.getOutPath() == "":
            self.printWarning("No output filepath given")
            return
        # make the directory
        checkAndMakeDir(self.getOutPath())
        # calculate our own cMap
        chanMap = {}
        for iChan in range(0, len(chanHeaders)):
            chanType = chanHeaders[iChan]["channel_type"]
            chanMap[chanType] = iChan
        # check if in physical units
        if physical:
            kwargs["scaling_applied"] = True
            self.dtype = np.float32
        # write the data
        self.write(headers, chanHeaders, chanMap, timeData, **kwargs)

    def write(
        self,
        headers: Dict,
        chanHeaders: List,
        chanMap: Dict,
        timeData: TimeData,
        **kwargs
    ):
        """Write out the header file

        Parameters
        ----------
        headers : Dict
            Dictionary of headers
        chanHeaders : List
            List of channel headers
        chanMap : Dict
            Maps channel to index for chanHeaders    
        timeData : TimeData
            Time series data as TimeData object        
        """

        # set global headers for keyword arguments
        headers = self.setGlobalHeadersFromKeywords(headers, kwargs)
        # set channel headers for keyword arguments
        chanHeaders = self.setChanHeadersFromKeywords(chanHeaders, kwargs)

        # now overwrite the options by checking the TimeData object
        # number of samples and sample frequency
        # Current method favours the time data object
        chans = sorted(list(timeData.chans))
        dataSizes = []
        for c in chans:
            dataSizes.append(timeData.data[c].size)
        if min(dataSizes) != max(dataSizes):
            self.printWarning(
                "Channels do not have the same number of samples: {} - {}".format(
                    ", ".join(chans), ", ".join(dataSizes)
                )
            )
            self.printWarning("Only the smallest number of samples will be written out")
        numSamples = min(dataSizes)
        if headers["num_samples"] != numSamples:
            self.printWarning(
                "Number of samples {} in headers does not match number of samples in TimeData object {}. TimeData info will be used.".format(
                    headers["num_samples"], numSamples
                )
            )
            headers["num_samples"] = numSamples
        timeData.numSamples = numSamples
        # sample freq
        if headers["sample_freq"] != timeData.sampleFreq:
            self.printWarning(
                "Sample frequency of {} Hz in headers does not match {} Hz in TimeData object".format(
                    headers["sample_freq"], timeData.sampleFreq
                )
            )
            self.printWarning("Sample frequency in TimeData object will be used")
            headers["sample_freq"] = timeData.sampleFreq

        # deal with start and end time and create datetime objects
        # the start time does not change on resampling, only the end time
        datetimeStart = datetime.strptime(
            "{} {}".format(headers["start_date"], headers["start_time"]),
            "%Y-%m-%d %H:%M:%S.%f",
        )
        datetimeStop = datetime.strptime(
            "{} {}".format(headers["stop_date"], headers["stop_time"]),
            "%Y-%m-%d %H:%M:%S.%f",
        )
        # now let's compare to the time data
        if datetimeStart != timeData.startTime:
            self.printWarning(
                "Start in headers {} does not match that in TimeData object {}. TimeData start time will be used".format(
                    datetimeStart, timeData.startTime
                )
            )
            datetimeStart = timeData.startTime
        if datetimeStop != timeData.stopTime:
            self.printWarning(
                "Stop in headers {} does not match that in TimeData object {}. TimeData stop time will be used".format(
                    datetimeStop, timeData.stopTime
                )
            )
            datetimeStop = timeData.stopTime
        # now recalculate datetime using the number of samples and compare again
        datetimeRecalc = self.calcStopDateTime(
            timeData.sampleFreq, numSamples, datetimeStart
        )
        if datetimeRecalc != datetimeStop:
            self.printWarning(
                "Note, discrepancy between stop time in given headers and those calculated from data"
            )
            self.printWarning(
                "Causes of this might be resampling or interpolation processes and the limiting of data"
            )
            self.printWarning(
                "If no resampling, interpolation or limiting of data has been performed, please check all times"
            )
            self.printWarning(
                "Stop time {} calculated from data will be used instead of that in data {}".format(
                    datetimeRecalc, datetimeStop
                )
            )
            datetimeStop = datetimeRecalc
        headers["start_date"] = datetimeStart.strftime("%Y-%m-%d")
        headers["start_time"] = datetimeStart.strftime("%H:%M:%S.%f")
        headers["stop_date"] = datetimeStop.strftime("%Y-%m-%d")
        headers["stop_time"] = datetimeStop.strftime("%H:%M:%S.%f")

        # now update all the chan headers and limit data to numSamples
        for c in chans:
            timeData.data[c] = timeData.data[c][:numSamples]
            cIndex = chanMap[c]
            chanHeaders[cIndex]["num_samples"] = headers["num_samples"]
            chanHeaders[cIndex]["sample_freq"] = headers["sample_freq"]
            chanHeaders[cIndex]["start_date"] = headers["start_date"]
            chanHeaders[cIndex]["start_time"] = headers["start_time"]
            chanHeaders[cIndex]["stop_date"] = headers["stop_date"]
            chanHeaders[cIndex]["stop_time"] = headers["stop_time"]

        # finally, check the number of measurement channels
        headers["meas_channels"] = len(chans)

        # now write out the headers and save to class variables
        self.writeHeaders(headers, chans, chanMap, chanHeaders)
        self.headers = headers
        self.chans = chans
        self.chanMap = chanMap
        self.chanHeaders = chanHeaders
        # write out comment file
        self.writeComments(timeData.comments)
        # write out the data files
        self.writeDataFiles(chans, timeData)

    def writeHeaders(
        self,
        headers: Dict[str, Any],
        chans: List[str],
        chanMap: Dict[str, int],
        chanHeaders: List[Dict],
        rename: bool = True,
    ) -> bool:
        """Write out the header file

        Parameters
        ----------
        headers : Dict
            Dictionary of headers
        chans : List[str]
            Channels as a list of strings
        chanMap : Dict
            Maps channel to index for chanHeaders
        chanHeaders : List
            List of channel headers
        rename : bool, optional
            Rename the output ats_data_files. Default is True and this is the case when writing out data which has been read in from a different source with pre-existing headers. However, if creating template header files, then set this to False.
        """

        # write out the global headers
        f = open(os.path.join(self.getOutPath(), "global.hdr"), "w")
        f.write("HEADER = GLOBAL\n")
        globalHeaderwords = self.globalHeaderwords()
        for gH in globalHeaderwords:
            f.write("{} = {}\n".format(gH, headers[gH]))
        f.close()

        # write out the channel headers
        chanHeaderwords = self.chanHeaderwords()
        for idx, c in enumerate(chans):
            cf = open(
                os.path.join(self.getOutPath(), "chan_{:02d}.hdr".format(idx)), "w"
            )
            cf.write("HEADER = CHANNEL\n")
            # use the chanMap to get the index of the chanHeaders list
            cIndex = chanMap[c]
            # change the data file if necessary
            if rename:
                chanHeaders[cIndex]["ats_data_file"] = "chan_{:02d}{}".format(
                    idx, self.extension
                )
            # write out all the header words
            for cH in chanHeaderwords:
                cf.write("{} = {}\n".format(cH, chanHeaders[cIndex][cH]))
            cf.close()
        return True

    def writeComments(self, comments: List[str]) -> None:
        """Write out a comments file

        Parameters
        ----------
        comments : List[str]
            List of strings with data comments
        """

        f = open(os.path.join(self.getOutPath(), "comments.txt"), "w")
        for c in comments:
            f.write("{}\n".format(c))
        # add another comment about writing
        f.write("Dataset written to {} on {}".format(self.getOutPath(), datetime.now()))
        f.close()

    def writeDataFiles(self, chans, timeData) -> None:
        """Write out data files"""

        raise NotImplementedError(
            "Write data files not implemented in base class. Only child classes should ever be instantiated."
        )

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        textLst.append("Output file path for data = {}".format(self.getOutPath()))
        # if it exists, print out the headers
        if self.headers:
            textLst.append("Global Headers")
            textLst.append(self.headers)
        # if exists, print out a list of chans
        if self.chans:
            textLst.append("Channels found:")
            textLst.append(self.chans)
        # if exists, print out the chanMap
        if self.chanMap:
            textLst.append("Channel Map")
            textLst.append(self.chanMap)
        # if it exists, print out the chanHeaders
        if self.chanHeaders:
            textLst.append("Channel Headers")
            for c in self.chans:
                textLst.append(c)
                textLst.append(self.chanHeaders[self.chanMap[c]])
        return textLst



import setuptools
import resistics

# options
ver = resistics.__version__

# See note below for more information about classifiers
classifiers = [
    "Development Status :: 4 - Beta",
    "Programming Language :: Python :: 3",
    "Operating System :: OS Independent",
    "License :: OSI Approved :: MIT License",
    "Intended Audience :: Science/Research",
]

with open("README.md", "r") as fh:
    long_description = fh.read()

project_urls = {
    "Source": "https://github.com/resistics/resistics",
    "Tracker": "https://github.com/resistics/resistics/issues",
}

dependencies = ["numpy", "scipy", "pyfftw", "matplotlib", "configobj"]

setuptools.setup(
    name="resistics",
    version=ver,
    description="Robust magnetotelluric processing package",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="geophysics magnetotellurics electromagnetic resistivity statistics",
    author="Neeraj Shah",
    author_email="resistics@outlook.com",
    url="https://www.resistics.io",
    project_urls=project_urls,
    license="MIT",
    python_requires=">=3.6",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=dependencies,
    classifiers=classifiers,
)




from os.path import basename, splitext
import numpy as np
import math
import collections
from typing import List, Dict, Union

# import from package
from resistics.ioHandlers.ioHandler import IOHandler
from resistics.dataObjects.transferFunctionData import TransferFunctionData


class TransferFunctionReader(IOHandler):
    """Class for reading and writing maskData

    Transfer function files are located in:
    project -> transFuncData -> site -> filepath

	Attributes
	----------
    filepath : str 
        Transfer function file
    headers : OrderedDict()
        Transfer function file header information

	Methods
	-------
	__init__(filepath="")
		Initialise the transferFunctionReader file
	read(filepath)
        Set and read a new transfer function file
    readFile()
        Read the transfer function file
    readInternal()
        Read internally formatted transfer function files
    formatHeaders()
        Format header values for internally formatted transfer function files
    readEdi()
        Read Edi formatted tranfer function files
    readNumericBlock(lines)
        Read a numeric block for Edi transfer function files
    constructComplex(real, imag)
        Construct a complex valued array from arrays of the real and imaginary values of the complex numbers
    printList()
        Class status returned as list of strings     
	"""

    def __init__(self, filepath: str = "") -> None:
        """Initialise

        Parameters
        ----------
        filepath : str, optional
            Path to transfer function file file
        """

        self.filepath: str = filepath
        self.headers = collections.OrderedDict()
        # read file if given
        if self.filepath != "":
            self.readFile()

    def read(self, filepath: str) -> None:
        """Read a new file

        Parameters
        ----------
        filepath : str
            Filepath to transfer function file
        """

        self.filepath = filepath
        self.readFile()

    def readFile(self) -> TransferFunctionData:
        """Read a transfer function file

        Returns
        -------
        TransferFunctionData
            Tranfser function data object
        """

        # get the extension and decide how to read it
        filename, extension = splitext(self.filepath)
        if "edi" in extension:
            return self.readEdi()
        else:
            return self.readInternal()

    def readInternal(self) -> None:
        """Read an internally formatted transfer function file

        Returns
        -------
        TransferFunctionData
            Tranfser function data object
        """

        key = ""
        impedances: Dict = {}
        variances: Dict = {}
        freq: Union[List[float], np.ndarray] = []
        with open(self.filepath, "r") as inFile:
            lines = inFile.readlines()
        numLines = len(lines)
        # find the headers
        for i in range(0, numLines):
            line = lines[i].strip()
            if line == "":
                continue
            if "Evaluation frequencies" not in line:
                split = line.split("=")
                self.headers[split[0].strip()] = split[1].strip()
            else:
                break
        # format headers
        self.formatHeaders()
        # read in the frequencies
        freq = np.loadtxt(lines[i + 1].strip().split(","), dtype="float")
        # read in the rest of the data
        for j in range(i + 2, numLines):
            line = lines[j].strip()
            if "Z" in line or "Var" in line:
                split = line.split("-")
                key = split[1].strip()
                j += 1
                if "Z" in line:
                    impedances[key] = np.loadtxt(
                        lines[j].strip().split(","), dtype="complex"
                    )
                if "Var" in line:
                    variances[key] = np.loadtxt(
                        lines[j].strip().split(","), dtype="float"
                    )
        self.tfData = TransferFunctionData(freq, impedances, variances)

    def formatHeaders(self) -> None:
        """Format  header values for the internal transfer function file format"""

        if "sampleFreq" in self.headers:
            self.headers["sampleFreq"] = float(self.headers["sampleFreq"])
        test = ["insite", "outsite", "remotesite"]
        for header in test:
            if header in self.headers and self.headers[header] == "False":
                self.headers[header] = False
        test = ["inchans", "outchans", "remotechans"]
        for header in test:
            if header in self.headers and self.headers[header] == "False":
                self.headers[header] = []

    def readEdi(self) -> None:
        """Read an EDI transfer function file

        EDI files have an >END that signifies the end of the file. This works quite nicely to our advantage when reading an EDI file.

        Returns
        -------
        TransferFunctionData
            Tranfser function data object
        """

        with open(self.filepath, "r") as inFile:
            lines = inFile.readlines()
        numLines = len(lines)
        section: Dict = {}
        sectionNum = 0
        sectionBreaks = []
        for i in range(0, numLines):
            # get the starting line of the sections
            # at the same time strip out all new lines characters
            lines[i] = lines[i].strip()
            if ">" in lines[i]:
                split = lines[i].split(" ")
                name = split[0][1:]
                # save the section
                section[name] = sectionNum
                # save the starting line of the section
                sectionBreaks.append(i)
                sectionNum = sectionNum + 1

        ## set headers
        startLine = sectionBreaks[section["HEAD"]]
        endLine = sectionBreaks[section["HEAD"] + 1]
        for i in range(startLine + 1, endLine):
            if lines[i] != "":
                split = lines[i].split("=")
                self.headers[split[0]] = split[1]

        ## set frequency information
        # get number of frequencies
        startLine = sectionBreaks[section["FREQ"]]
        endLine = sectionBreaks[section["FREQ"] + 1]
        split = lines[startLine].split(" ")
        numFreq = 0
        for s in split:
            if "NFREQ" in s:
                split2 = s.split("=")
                numFreq = int(split2[1])
        freq = self.readNumericBlock(lines[startLine + 1 : endLine])

        ## read impedance tensor components and variances
        components = ["ZXX", "ZYY", "ZXY", "ZYX"]
        polarisations = ["ExHx", "EyHy", "ExHy", "EyHx"]
        variances = {}
        impedances = {}

        for comp, pol in zip(components, polarisations):
            # real
            realComp = "{}R".format(comp)
            startLine = sectionBreaks[section[realComp]]
            endLine = sectionBreaks[section[realComp] + 1]
            real = self.readNumericBlock(lines[startLine + 1 : endLine])
            # imag
            imagComp = "{}I".format(comp)
            startLine = sectionBreaks[section[imagComp]]
            endLine = sectionBreaks[section[imagComp] + 1]
            imag = self.readNumericBlock(lines[startLine + 1 : endLine])
            # impedance
            impedances[pol] = self.constructComplex(real, imag)
            # variance
            varComp = "{}.VAR".format(comp)
            startLine = sectionBreaks[section[varComp]]
            endLine = sectionBreaks[section[varComp] + 1]
            variances[pol] = self.readNumericBlock(lines[startLine + 1 : endLine])

        # initialise impedance tensor
        self.tfData = TransferFunctionData(freq, impedances, variances)

    def readNumericBlock(self, lines: List[str]):
        """Read an EDI numeric block

        Parameters
        ----------
        lines : List[str]
            List of strings (or the lines in the file) that are a numeric block

        Returns
        -------
        np.ndarray
            Numeric data as a numpy array
        """

        tmp = []
        size = len(lines)
        for i in range(0, size):
            if lines[i] == "":
                continue
            split = lines[i].split(" ")
            # add the numbers
            tmp = tmp + split
        # convert to float
        arr = np.zeros(shape=(len(tmp)))
        for idx, t in enumerate(tmp):
            arr[idx] = float(t)
        return arr

    def constructComplex(self, real, imag):
        """Construct a complex array from separate real and imaginary arrays 

        Parameters
        ----------
        real : np.ndarray
            An array of the real part of the complex data
        imag : np.ndarray
            An array of the complex part of the complex data

        Returns
        -------
        np.ndarray
            Numpy array of complex data
        """

        size = len(real)
        arr = np.zeros(shape=(size), dtype="complex")
        for i in range(0, size):
            arr[i] = complex(real[i], imag[i])
        return arr

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        textLst.append("Headers in file")
        for h, val in self.headers.items():
            textLst.append("\t{} = {}".format(h, val))
        return textLst




import os
import glob
from datetime import datetime, timedelta
import numpy as np
from typing import List, Tuple

# import from package
from resistics.ioHandlers.dataReader import DataReader
from resistics.utilities.utilsIO import checkFilepath, lineToKeyAndValue


class DataReaderInternal(DataReader):
    """Data reader for internal formatted data

    Internal formatted data is straightforward. Header information is read in from a global header file and header files for each channel, all of them ascii formatted file. Each channel has its own data file written out using numpy's binary write function and with a .dat extension.

    As raw data is not usually in the internal data format, to avoid any problems, the following workflow is suggested:

    - Get physical data from the raw data files (ATS, SPAM, Phoenix)
    - Perform any pre-processing required
    - Save as internal format, ensuring to set the scaling_applied channel header for every channel to True. This will avoid any further scaling when the data is read in again.
    
    Methods
    -------
    dataHeaders()
        Headers to read in
    readHeaders()
        Specific function for reading the headers for internal format
    lineToKeyAndValue(line)
        Separate a line into key and value with = as a delimiter
    """

    def dataHeaders(self) -> Tuple[List[str], List[str], List[str], List[str]]:
        """Return the data headers in the internal file format
        
        Returns
        -------
        recordingHeaders : List[str]
            Headers with information about the recording
        globalHeaders : List[str]
            Common headers with information about the recording
        channelHeadersInput : List[str]
            Channel setup headers
        channelHeadersOutput : List[str]
            Channel recording headers      
        """

        # note, in comparison to ats headers, this also has one called scaling_applied
        recordingHeaders = ["start_time", "start_date", "stop_time", "stop_date"]
        globalHeaders = ["meas_channels", "sample_freq"]
        channelHeadersInput = ["gain_stage1", "gain_stage2", "hchopper", "echopper"]
        channelHeadersOutput = [
            "start_time",
            "start_date",
            "sample_freq",
            "num_samples",
            "ats_data_file",
            "sensor_type",
            "channel_type",
            "ts_lsb",
            "scaling_applied",
            "pos_x1",
            "pos_x2",
            "pos_y1",
            "pos_y2",
            "pos_z1",
            "pos_z2",
            "sensor_sernum",
        ]
        return (
            recordingHeaders,
            globalHeaders,
            channelHeadersInput,
            channelHeadersOutput,
        )

    def readHeader(self) -> None:
        """Read time data header file for internal format"""

        # first read the global headers
        # look in headerF for global.hdr
        if os.path.join(self.dataPath, "global.hdr") not in self.headerF:
            self.printError(
                "Global header not found. The global.hdr file is required", quitRun=True
            )
        globalF = open(os.path.join(self.dataPath, "global.hdr"))
        lines = globalF.readlines()
        globalF.close()
        # ignore the first line
        lines.pop(0)
        # now go through and get the headers
        self.headers = {}
        for l in lines:
            if l == "":
                continue
            key, val = lineToKeyAndValue(l.strip())
            self.headers[key] = val

        # now want to deal with the chan headers
        numChans = int(self.headers["meas_channels"])
        self.chanHeaders = []
        for iChan in range(0, numChans):
            chanF = open(os.path.join(self.dataPath, "chan_{:02d}.hdr".format(iChan)))
            lines = chanF.readlines()
            chanF.close()
            # remove first line and read the headers for the channel
            lines.pop(0)
            cHeader = {}
            for l in lines:
                if l == "":
                    continue
                key, val = lineToKeyAndValue(l.strip())
                cHeader[key] = val
            self.chanHeaders.append(cHeader)

        # finally, read the comments
        commentPath = os.path.join(self.dataPath, "comments.txt")
        if checkFilepath(commentPath):
            f = open(commentPath, "r")
            self.comments = f.readlines()
            for idx, comment in enumerate(self.comments):
                self.comments[idx] = comment.rstrip()
            f.close()
        else:
            self.comments = []


import os
import numpy as np
import math
from datetime import datetime, timedelta
from typing import List, Tuple

# import from package
from resistics.ioHandlers.ioHandler import IOHandler
from resistics.dataObjects.maskData import MaskData
from resistics.utilities.utilsIO import checkAndMakeDir, fileFormatSampleFreq
from resistics.utilities.utilsPrint import listToString


class MaskIO(IOHandler):
    """Class for reading and writing maskData

    Masks are referenced to sampling frequencies rather than particular measurements. The idea if that masks are calculated out for all the data and later, the data can be constrained either using date and time constraints or by using window masks based on statistics.

    Mask files are located in:
    project -> maskData -> site -> sample frequency -> specdir

	Attributes
	----------
	datapath : str
		Path to mask file directory

	Methods
	-------
	__init__(datapath)
		Initialise the MaskIO object.
	read(maskName, sampleFreq)
		Read in maskName for sampleFreq from datapath        
	write(maskData)
		Write out maskData object to datapath.
	getFileNames(makeName, sampleFreq)
		The filename for a maskFile given a maskName and sampling frequency
    printList()
        Class status returned as list of strings       
	"""

    def __init__(self, datapath: str = "") -> None:
        """Initialise

        Parameters
        ----------
        datapath : str, optional
            Path to mask file directory
        """

        self.datapath: str = datapath

    def read(self, maskName: str, sampleFreq: float) -> MaskData:
        """Read in maskData from a file defined by maskName and sampleFreq

        Parameters
        ----------
        maskName : MaskData
            MaskData object
        sampleFreq : float
            The sampling frequency of the data

        Returns
        -------
        maskData : MaskData
            The MaskData object
        """

        # read the window file
        infoName, winName = self.getFileNames(maskName, sampleFreq)
        infoFile = open(infoName, "r")
        lines = infoFile.readlines()
        infoFile.close()
        # this is all passed into the constructor
        sampleFreq = float(lines[0].strip())
        numLevels = int(lines[1].strip())
        evalFreq = []
        for iL in range(0, numLevels):
            evalFreq.append(
                list(np.fromstring(lines[2 + iL].strip(), dtype=float, sep=","))
            )
        # read in the stats
        stats = lines[2 + numLevels].strip().split(",")
        for idx, s in enumerate(stats):
            stats[idx] = stats[idx].strip()
        # now create a MaskData object
        maskData = MaskData(sampleFreq, numLevels, evalFreq, name=maskName, stats=stats)
        lines = lines[3 + numLevels :]
        # now read in the statistics
        evalFreqSorted = sorted(list(maskData.constraints.keys()))
        eIdx = -1
        for l in lines:
            if "Frequency" in l:
                eIdx = eIdx + 1
            elif "Statistic" in l:
                stat = l.strip().split("=")[1]
                stat = stat.strip()
            else:
                split = (
                    l.strip().split()
                )  # component is 0, min is 1, max is 2, inout is 3
                maskData.constraints[evalFreqSorted[eIdx]][stat][split[0]] = [
                    float(split[1]),
                    float(split[2]),
                ]
                inOut = False
                if split[3] == "True":
                    inOut = True
                maskData.insideOut[evalFreqSorted[eIdx]][stat][split[0]] = inOut
        # now read the window mask file
        winMaskArray = np.load(winName + ".npy")
        for eIdx, eFreq in enumerate(evalFreqSorted):
            maskData.maskWindows[eFreq] = set(winMaskArray[eIdx])
            # remove -1 from the set
            maskData.maskWindows[eFreq] = maskData.maskWindows[eFreq] - set([-1])
        # now want to return as MaskData object
        return maskData

    def write(self, maskData: MaskData) -> None:
        """Write the maskData out to datapath

        Mask data is saved as a numpy binary object

        Parameters
        ----------
        maskData : MaskData
            MaskData object
        """

        infoName, winName = self.getFileNames(maskData.maskName, maskData.sampleFreq)
        infoFile = open(infoName, "w")
        # first write out constraints
        infoFile.write("{:.9f}\n".format(maskData.sampleFreq))
        infoFile.write("{}\n".format(maskData.numLevels))
        for iL in range(0, maskData.numLevels):
            infoFile.write("{}\n".format(listToString(maskData.evalFreq[iL])))
        infoFile.write("{}\n".format(", ".join(maskData.stats)))
        # now write out the data file
        # first get a sorted list of all the evaluations frequencies to loop through
        evalFreq = sorted(list(maskData.constraints.keys()))
        for eFreq in evalFreq:
            infoFile.write("Frequency = {:.9f}\n".format(eFreq))
            for stat in maskData.stats:
                infoFile.write("Statistic = {}\n".format(stat))
                for component in maskData.constraints[eFreq][stat]:
                    minVal = maskData.constraints[eFreq][stat][component][0]
                    maxVal = maskData.constraints[eFreq][stat][component][1]
                    infoFile.write(
                        "{}\t{}\t{}\t{}\n".format(
                            component,
                            minVal,
                            maxVal,
                            maskData.insideOut[eFreq][stat][component],
                        )
                    )
        # then loop through each
        infoFile.close()
        maskSize = 0
        for eFreq in evalFreq:
            if len(maskData.maskWindows[eFreq]) > maskSize:
                maskSize = len(maskData.maskWindows[eFreq])
        # create window mask array and initalise to -1
        winMaskArray = np.ones(shape=(len(evalFreq), maskSize), dtype=int) * -1
        # now fill the array
        for eIdx, eFreq in enumerate(evalFreq):
            lst = list(maskData.maskWindows[eFreq])
            winMaskArray[eIdx, 0 : len(lst)] = lst
        np.save(winName, winMaskArray)

    def getFileNames(self, maskName: str, sampleFreq: float) -> Tuple[str, str]:
        """Get the name of a mask file
        
        This method is here to give consistent file names for mask files.

        Parameters
        ----------
        maskName : str
            The name of the mask
        sampleFreq : float
            The sampling frequency of the data

        Returns
        -------
        infoFile : str
            The name of the mask infoFile
        winFile : str
            The name of the mask winFile
        """

        checkAndMakeDir(self.datapath)
        sampleFreqStr = fileFormatSampleFreq(sampleFreq)
        name: str = maskName + "_{}".format(sampleFreqStr)
        infoFile: str = os.path.join(self.datapath, name + ".info")
        winFile: str = os.path.join(
            self.datapath, name
        )  # no need for extension here, numpy adds one
        return infoFile, winFile

    def printList(self) -> List[str]:
        """Class information as a list of strings

        Returns
        -------
        out : List[str]
            List of strings with information
        """

        textLst = []
        if self.datapath == "":
            textLst.append(
                "No datapath given. Please set the datapath attribute of the class"
            )
        else:
            textLst.append("Datapath = {}".format(self.datapath))

        return textLst

